import math
import anndata
import numpy as np
import numba as nb
import pandas as pd
from numba import prange, njit
from anndata import AnnData
from scipy.stats.mstats import gmean
import copy
from typing import Optional, Union, Tuple, Literal, List
from ..utils import _scale, _graph, _add_info_from_sample, _check_adata_type
from tqdm import tqdm
from .utils import adj_pvals, _count_edge, allocation_edge_2_diff_cell_type
from ..pp.utils import _preprocessing_of_graph
import logging as logg
import re
import os

__all__ = ['cell_level_communications', 'cell_type_level_communication', 'lr_pairs']


class lr_pairs():
    """
    Class for ligand and receptor pairs
    """

    def __init__(
            self,
            lr_data: Union[pd.DataFrame, Literal['human', 'mouse']] = 'human',
            Annotation_key: Optional[str] = 'annotation',
            ligand_key: Optional[str] = 'ligand_symbol',
            receptor_key: Optional[str] = 'receptor_symbol',
            ):
        """
        Parameters
        ----------
        lr_data : pd.DataFrame
            Ligand and receptor information database. Default: CellChat human database
        Annotation_key : str
            The key of Annotation (Contact or secretory) in lr_data
        ligand_key : str
            The key of ligand in lr_data
        receptor_key : str
            The key of receptor in lr_data
        """

        self.ligand_key = ligand_key
        self.receptor_key = receptor_key
        self.Annotation_key = Annotation_key

        if lr_data is 'human':
            path = os.path.dirname(os.path.realpath(__file__))
            lr_data = pd.read_csv(f'{path}/datasets/Human-2020-Jin-LR-pairs.csv', index_col=0, header=0)
            lr_data = lr_data.replace('ECM-Receptor', 'Cell-Cell Contact')
        elif lr_data is 'mouse':
            path = os.path.dirname(os.path.realpath(__file__))
            lr_data = pd.read_csv(f'{path}/datasets/Mouse-2020-Jin-LR-pairs.csv', index_col=0, header=0)
            lr_data = lr_data.replace('ECM-Receptor', 'Cell-Cell Contact')

        if Annotation_key is None:
            self.lr_data = lr_data.loc[:, [ligand_key, receptor_key]]
        else:
            self.lr_data = lr_data.loc[:, [ligand_key, receptor_key, Annotation_key]]

        self.complexes = {}
        self.gene_name = []

    def get_complexes(self,
                      complex_sep='&',
                      ):
        """
        Get the names of all the complexes

        Parameters
        ----------
        complex_sep
            A separator used to separate subunits in the database, default: '&'

        Returns
        -------
        self.complexes
            A dict for complexes name and its subunits (key: complexes name, value: subunits)
        self.gene_name
            A list for all genes' name
        """

        col_a = self.ligand_key
        col_b = self.receptor_key
        complexes = {}
        gene_name = []

        for idx, row in self.lr_data.iterrows():
            prot_a = row[col_a]
            prot_b = row[col_b]

            gene_name.append(prot_a)
            gene_name.append(prot_b)

            if complex_sep in prot_a:
                comp = set([l for l in prot_a.split(complex_sep)])
                complexes[prot_a] = comp
            if complex_sep in prot_b:
                comp = set([r for r in prot_b.split(complex_sep)])
                complexes[prot_b] = comp

        self.complexes = complexes
        self.gene_name = list(set(gene_name))


def get_used_lr(
        lr_pairs: lr_pairs,
        key=None
) -> tuple[list, list]:
    """
    Collate and return information of ligand-receptor pairs
    """
    lr_names = []
    lr_used = []
    ln = lr_pairs.ligand_key
    rn = lr_pairs.receptor_key
    if key is None:
        for id, row in lr_pairs.lr_data:
            lr_names.append(row[ln] + ':' + row[rn])
            lr_used.append([row[ln], row[rn]])
    else:
        for id, row in lr_pairs.lr_data.iterrows():
            if row[lr_pairs.Annotation_key] == key:
                lr_names.append(row[ln] + ':' + row[rn])
                lr_used.append([row[ln], row[rn]])
    return lr_names, lr_used


def _lr_helper(
        adata: anndata.AnnData,
        lr_pairs: lr_pairs,
        scale: Union[str, float],
        key: str,
        method: str,
        norm: bool = True,
        percent_of_drop: Union[int, float] = 0,
        radius: Optional[float] = None,
        max_quantile: Optional[float] = 98,
        agg_method: str = 'gmean',
        spatial_in_obsm: str = 'spatial',
):
    """
    Data pre-processing
    """
    # Get spatial information
    df_pixel = adata.obsm[spatial_in_obsm]
    if isinstance(df_pixel, np.ndarray):
        df_pixel = pd.DataFrame(df_pixel)
        df_pixel = df_pixel.iloc[:, [0, 1]]
    row = list(df_pixel.iloc[:, 1] * scale)
    col = list(df_pixel.iloc[:, 0] * scale)

    indices, distances = _graph(
        col, row,
        method=method,
        cutoff=radius,
        max_quantile=max_quantile,
    )

    # lr-pairs information
    lr_names, lr_used = get_used_lr(lr_pairs, key)
    if not isinstance(adata.X, np.ndarray):
        exp = adata.X.toarray()
    else:
        exp = adata.X
    dict_gene = {gene: index for index, gene in enumerate(adata.var_names)}
    if lr_pairs.complexes is not None:
        exp, dict_gene = update_exp(exp,
                                    dict_gene,
                                    lr_pairs.complexes,
                                    agg_method,
                                    lr_pairs.gene_name,
                                    norm,
                                    percent_of_drop,
                                    )

    return indices, distances, lr_names, lr_used, exp.astype(np.float32), dict_gene


def update_exp(
        exp: np.ndarray,
        dict_gene: dict,
        complexes: dict,
        agg_method: str,
        used_gene: list,
        norm: bool,
        percent_of_drop: Union[int, float]
):
    """
    Update expression by combining subunit from database. if there are 'ligand1 : receptor1 and receptor2' in dataset.
     'receptor1 and receptor2' expression will be generated by the amount of 'receptor1' and 'receptor2' expression

    Parameters
    ----------
    exp : np.ndarray
        The expression of all the ligand and receptor in anndata.
    dict_gene : dict
        The dict of ligand and receptor.
    complexes : dict
        The dict of complexes and their subunits.
    agg_method : str
        Method for calculating complex expression. Default: geometric mean.
    used_gene : list
        All the genes that are used.
    norm : bool
        Whether 0-1 normalization was performed on the expression volume data.
    percent_of_drop : Union[int, float]
        Percentage of extreme values removed from normalization

    Returns
    -------
    exp_new
        The expression after the addition of complex
    dict_gene_new
        The dict_gene after the addition of complex

    """
    gene_name = list(dict_gene.keys())
    exp = copy.deepcopy(exp)
    for k, v in complexes.items():
        if all(g in gene_name for g in v):
            dict_gene[k] = len(dict_gene)
            index = [dict_gene[gene] for gene in v]
            if agg_method == 'min':
                update_ = np.min(exp[:, index], axis=1)
            elif agg_method == 'mean':
                update_ = np.mean(exp[:, index], axis=1)
            elif agg_method == 'gmean':
                update_ = gmean(exp[:, index], axis=1)
            else:
                ValueError("{} is not a valid agg_method".format(agg_method))
            exp = np.insert(exp, exp.shape[1], np.array(update_), axis=1)

    dict_gene_new = {}
    i = 0
    index_used = []
    for gene in list(set(used_gene)):
        if gene in dict_gene.keys():
            index_used.append(dict_gene[gene])
            dict_gene_new[gene] = i
            i += 1
    exp_new = exp[:, index_used]

    if norm:
        exp_new = ((exp_new - np.percentile(exp_new, percent_of_drop, axis=0)) /
                  (np.percentile(exp_new, 100 - percent_of_drop, axis=0) - np.percentile(exp_new, percent_of_drop, axis=0)))
        exp_new = np.where(exp_new > 1, 1.0, exp_new)
        exp_new = np.where(exp_new < 0, 0.0, exp_new)
        exp_new = np.nan_to_num(exp_new)

    return exp_new.astype(np.float32), dict_gene_new


def get_sample_lr_score(
        lr_used: list,
        species: str,
        exp: np.ndarray,
        dict_gene: dict,
        gene_used: list,
        indices: list,
        distances: list,
        lr_num: int,
        obs_num: int,
        func: bool,
        n_shuffle: int,
):
    """
    All communication scores for a certain communication type
    """
    if func:
        mode = 'contact'
    else:
        mode = 'secretory'

    without_neighbor = 0
    neigh_num = 0
    i=1
    for i, j in enumerate(indices):
        if not isinstance(j, np.ndarray):
            indices[i] = np.array([], dtype=np.int32)
            distances[i] = np.array([], dtype=np.float32)
            without_neighbor += 1
        else:
            indices[i] = np.array(indices[i], dtype=np.int32)
            distances[i] = np.array(distances[i], dtype=np.float32)
        neigh_num += len(indices[i])
    print(f'In {mode} mode, The average number of neighbors is {neigh_num/i}')
    print(f'In {mode} mode, total of {without_neighbor} spots have no neighbors')

    indices = nb.typed.List(indices)
    distances = nb.typed.List(distances)
    spots_score = np.zeros(shape=(obs_num, lr_num), dtype=np.float32)
    spots_p = np.ones(shape=(obs_num, lr_num), dtype=np.float32) * (n_shuffle + 1)

    with tqdm(
            total=lr_num,
            desc=f"{lr_num} {mode} ligand-receptor pairs.",
            bar_format="{l_bar}{bar} [ time left: {remaining} ]",
    ) as pbar:

        for index_lr in prange(lr_num):
            ligand, receptor = str(lr_used[index_lr][0]), str(lr_used[index_lr][1])
            if ligand not in gene_used or receptor not in gene_used:
                spots_score[:, index_lr] = 0
                pbar.update(1)
                continue

            l_index = int(dict_gene[ligand])
            r_index = int(dict_gene[receptor])
            if species == 'ligand':
                exp_index = exp[:, l_index]
                exp_neighbor = exp[:, r_index]
            else:
                exp_index = exp[:, l_index]
                exp_neighbor = exp[:, r_index]

            spots_score[:, index_lr], spots_p[:, index_lr] = get_one_lr_score(
                obs_num=obs_num,
                exp_index=exp_index,
                exp_neighbor=exp_neighbor,
                neighbors=indices,
                distances=distances,
                n=n_shuffle,
                func=func
            )
            pbar.update(1)

    return spots_score, spots_p / (n_shuffle + 1)


@njit()
def get_one_lr_score(
        obs_num: int,
        exp_index: np.ndarray,
        exp_neighbor: np.ndarray,
        neighbors: np.ndarray,
        distances: np.ndarray,
        n: int,
        func: bool,
):
    """
    Compute the cell-cell communication score for a pair of ligand receptors in spot level.
    """
    scores_all_spots = np.zeros(shape=obs_num, dtype=np.float32)
    p_all_spots = np.ones(shape=obs_num, dtype=np.int32) * (n + 1)
    exps_index = np.zeros(shape=(exp_index.shape[0], n + 1), dtype=np.float32)
    exps_neigh = np.zeros(shape=(exp_neighbor.shape[0], n + 1), dtype=np.float32)
    exps_index[:, 0] = exp_index
    exps_neigh[:, 0] = exp_neighbor

    for index in range(n):
        np.random.shuffle(exp_index)
        np.random.shuffle(exp_neighbor)
        exps_index[:, index + 1] = exp_index
        exps_neigh[:, index + 1] = exp_neighbor

    for index_spot in prange(obs_num):
        neighbor = neighbors[index_spot]
        if len(neighbor) == 0:
            continue
        else:
            if func:
                lr_score = np.sum(exps_index[index_spot, :] * exps_neigh[neighbor, :], axis=0, dtype=np.float32)
            else:
                d = distances[index_spot]
                d_reshaped = np.reshape(d + np.float32(1.0), (d.shape[0], 1))
                lr_score = np.sum(exps_index[index_spot, :] * (exps_neigh[neighbor, :] / d_reshaped), axis=0,
                                  dtype=np.float32)
            scores_all_spots[index_spot] = lr_score[0]

            count_shuffle = 0
            score_true = lr_score[0]
            for score in lr_score:
                if score >= score_true:
                    count_shuffle += 1
            p_all_spots[index_spot] = count_shuffle

    return scores_all_spots, p_all_spots


def get_contact_score(
        adata: AnnData,
        lr_pairs,
        species,
        key='contact',
        scale=1.0,
        agg_method='gmean',
        norm: bool = True,
        percent_of_drop: Union[int, float] = 0,
        spatial_in_obsm='spatial',
        max_quantile=95,
        n_iters: int = 1000,
):
    """
    contact type communication score in spot level
    """

    # if type(n_jobs) != type(None):
    #     nb.set_num_threads(n_jobs)

    indices, distances, lr_names, lr_used, exp, dict_gene = _lr_helper(adata=adata,
                                                                       lr_pairs=lr_pairs,
                                                                       scale=scale,
                                                                       key=key,
                                                                       method='neighbor',
                                                                       max_quantile=max_quantile,
                                                                       agg_method=agg_method,
                                                                       norm=norm,
                                                                       percent_of_drop=percent_of_drop,
                                                                       spatial_in_obsm=spatial_in_obsm,
                                                                       )

    print(f'There are {len(lr_names)} cell-cell contact ligand-receptor pairs in sample')

    obs_names = adata.obs_names.tolist()
    lr_score, pvalue = get_sample_lr_score(
        lr_used=lr_used,
        species=species,
        exp=exp,
        dict_gene=dict_gene,
        gene_used=list(dict_gene.keys()),
        indices=indices,
        distances=distances,
        lr_num=len(lr_used),
        obs_num=len(obs_names),
        func=True,
        n_shuffle=n_iters
    )

    return lr_score, pvalue, lr_names


def get_secretory_score(
        adata: AnnData,
        lr_pairs,
        species,
        key='secretory',
        scale=1.0,
        agg_method='gmean',
        norm: bool = True,
        percent_of_drop: Union[int, float] = 0,
        spatial_in_obsm='spatial',
        radius=100,
        n_iters: int = 1000,
):
    """
    secretory type communication score in spot level
    """

    # if type(n_jobs) != type(None):
    #     nb.set_num_threads(n_jobs)

    indices, distances, lr_names, lr_used, exp, dict_gene = _lr_helper(adata=adata,
                                                                       lr_pairs=lr_pairs,
                                                                       scale=scale,
                                                                       key=key,
                                                                       method='radius',
                                                                       radius=radius,
                                                                       max_quantile=None,
                                                                       agg_method=agg_method,
                                                                       norm=norm,
                                                                       percent_of_drop=percent_of_drop,
                                                                       spatial_in_obsm=spatial_in_obsm,
                                                                       )

    print(f'There are {len(lr_names)} secreted ligand-receptor pairs in sample')

    obs_names = adata.obs_names.tolist()
    lr_score, pvalue = get_sample_lr_score(
        lr_used=lr_used,
        species=species,
        exp=exp,
        dict_gene=dict_gene,
        gene_used=list(dict_gene.keys()),
        indices=indices,
        distances=distances,
        lr_num=len(lr_used),
        obs_num=len(obs_names),
        func=False,
        n_shuffle=n_iters
    )

    return lr_score, pvalue, lr_names


def cell_level_communications(
        adata: AnnData,
        lr_pairs: lr_pairs,
        species: Literal['ligand', 'receptor'] = 'ligand',
        agg_method: Literal['min', 'mean', 'gmean'] = 'gmean',
        scale: Union[str, float] = 'hires',
        sample: Union[str, int, list, None] = None,
        sample_key: Optional[str] = None,
        contact_key: str = 'contact',
        contact_max_quantile: int = 95,
        secretory_key: str = 'secretory',
        secretory_radius: Union[int, float] = 100,
        fdr_axis: Literal['spot', 'lr'] = 'spot',
        pval_adj_cutoff: float = 0.05,
        adj_method: str = 'fdr_bh',
        norm: bool = True,
        percent_of_drop: Union[int, float] = 0,
        spatial_in_obsm: str = 'spatial',
        n_iters: int = 1000,
        inplace: bool = True,
) -> anndata.AnnData:
    """
    A permutation test of ligand-receptor expression across every spot.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    lr_pairs : lr_pairs
        database of ligand-receptor pairs.
    species : Literal['ligand', 'receptor'], optional
        'ligand': The central cell is the ligand cell and the neighbor cells are the receptor cells.
        'receptor': The central cell is the receptor cell and the neighbor cells are the ligand cells.
    agg_method : str, optional
        Integrated subunit to calculate ligand or receptor.
        'min': ligand or receptor = min(sub1, sub2, sub3, ...)
        'mean': ligand or receptor = mean(sub1, sub2, sub3, ...)
        'gmean': geometrical mean, ligand or receptor = np.gmean(sub1, sub2, sub3, ...)
    scale : Union[str, float], optional
        scale used in subsequent analyses. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this
    sample : Union[str, int, list], optional
        Samples for which communication scores need to be calculated
    sample_key : str, optional
        The keyword of sample id in adata.obs.columns.
    contact_key : str, optional
        The tag name that represents the contact type in the LR database
    contact_max_quantile : int
        In Neighbor network, Order the distance of all sides, and more than max_quantile% of the sides will be removed
    secretory_key : str, optional
        The tag name that represents the secretory type in the LR database
    secretory_radius : Union[int, float]
        The maximum distance considered for the secretory type
    fdr_axis : Literal['spot', 'lr']
        Dimensions of applying multiple hypothesis testing
        'spot': tested in each spot
        'lr': tested in each ligand-receptor pair
        None: no multiple hypothesis testing
    pval_adj_cutoff : float
        Cutoff for spot to be significant based on adjusted p-value.
    adj_method : str
        Any method supported by statsmodels.stats.multitest.multipletests;
        https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html
    norm : bool
        Whether 0-1 normalization was performed on the expression volume data.
    percent_of_drop : Union[int, float]
        Percentage of extreme values removed from normalization
    spatial_in_obsm : str
        The key of spatial coordinates in adata.obsm
    n_iters : int
        Number of permutations for the permutation-test
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    - :attr:`anndata.AnnData.uns` ``['SOAPy']['ligand_spot_comm_score' or 'receptor_spot_comm_score']``
        - 'contact_names': Names of used contact type lR pairs.
        - 'secretory_names': Names of used secretory type lR pairs.
        - 'contact_affinity': The affinity of contact type lR pairs.
        - 'secretory_affinity': The affinity of secretory type lR pairs.

    """
    adata = _check_adata_type(adata, spatial_in_obsm, inplace)

    if isinstance(sample, str):
        sample = [sample]
    if sample is not None and sample_key is None:
        logg.error(f'Mult-sample niche analysis cannot be specified without a given sample key')
        raise ValueError
    if sample is None and sample_key is not None:
        sample = list(np.unique(adata.obs[sample_key]))
        logg.info(f'Use all samples in the niche calculation')
    if sample is None and sample_key is None:
        sample = [None]

    for sample_id in sample:
        if sample_id is None:
            bdata = adata
        else:
            bdata = adata[adata.obs[sample_key] == sample, :]

        if type(scale) != float:
            scale = _scale(bdata, None, scale)

        if contact_key is None:
            lr_score_c, pvalue_c, contact_names = None, None, None
            update_p_data_c = None
        else:
            lr_score_c, pvalue_c, contact_names = get_contact_score(
                adata=bdata,
                lr_pairs=lr_pairs,
                species=species,
                key=contact_key,
                scale=scale,
                agg_method=agg_method,
                spatial_in_obsm=spatial_in_obsm,
                max_quantile=contact_max_quantile,
                norm=norm,
                percent_of_drop=percent_of_drop,
                n_iters=n_iters,
            )
            update_p_data_c = adj_pvals(lr_score_c, pvalue_c, pval_adj_cutoff, fdr_axis, adj_method)

        if secretory_key is None:
            lr_score_s, pvalue_s, secretory_names = None, None, None
            update_p_data_s = None
        else:
            lr_score_s, pvalue_s, secretory_names = get_secretory_score(
                adata=bdata,
                lr_pairs=lr_pairs,
                species=species,
                key=secretory_key,
                scale=scale,
                agg_method=agg_method,
                spatial_in_obsm=spatial_in_obsm,
                radius=secretory_radius,
                norm=norm,
                percent_of_drop=percent_of_drop,
                n_iters=n_iters,
            )
            update_p_data_s = adj_pvals(lr_score_s, pvalue_s, pval_adj_cutoff, fdr_axis, adj_method)

        communication_ = {
            'contact_names': contact_names,
            'secretory_names': secretory_names,
            'contact_affinity': update_p_data_c,
            'secretory_affinity': update_p_data_s,
        }

        _add_info_from_sample(adata, sample_id=sample_id, keys=species+'_cell_comm_score', add=communication_)

    return adata


###################################
###################################
@njit()
def get_celltype_score(
        exp_ligand, exp_receptor, len_celltype, clusters, edges, distances, func, n_iters,
):
    """
    Compute the cell-cell communication score for a pair of ligand receptors in cell type level.
    """

    lr_score = np.zeros(shape=(n_iters, len_celltype, len_celltype), dtype=np.float32)
    lr_p = np.zeros(shape=(len_celltype, len_celltype), dtype=np.float32)
    shuffle_ligand = np.copy(exp_ligand)
    shuffle_receptor = np.copy(exp_receptor)
    # index = np.arange(len(clusters), dtype=np.int32)

    for i in range(n_iters):
        if i != 0:

            clusters_i = np.copy(clusters)
            shuffle_ligand_i = np.zeros_like(shuffle_ligand)
            shuffle_receptor_i = np.zeros_like(shuffle_receptor)

            for j in range(len_celltype):
                index_j = np.random.permutation(np.where(clusters == j)[0])
                shuffle_ligand_i[np.where(clusters == j)[0]] = shuffle_ligand[index_j]
                shuffle_receptor_i[np.where(clusters == j)[0]] = shuffle_receptor[index_j]

        else:
            clusters_i = np.copy(clusters)
            shuffle_ligand_i = np.copy(shuffle_ligand)
            shuffle_receptor_i = np.copy(shuffle_receptor)
        lr_score_i = np.zeros(shape=(len_celltype, len_celltype), dtype=np.float32)

        for index_edge in range(edges.shape[0]):
            spoti, spotj = edges[index_edge, 0], edges[index_edge, 1]
            d = distances[index_edge]
            cti = clusters_i[spoti]
            ctj = clusters_i[spotj]
            if func:
                lr_score_i[cti, ctj] += shuffle_ligand_i[spoti] * shuffle_receptor_i[spotj]
                lr_score_i[ctj, cti] += shuffle_ligand_i[spotj] * shuffle_receptor_i[spoti]
            else:
                lr_score_i[cti, ctj] += shuffle_ligand_i[spoti] * shuffle_receptor_i[spotj] / (
                            d + np.float32(1.0))
                lr_score_i[ctj, cti] += shuffle_ligand_i[spotj] * shuffle_receptor_i[spoti] / (
                            d + np.float32(1.0))

        lr_score[i, :, :] = lr_score_i

    lr_score_truth = lr_score[0, :, :]
    for i in range(n_iters):
        p = np.where(lr_score[i, :, :] >= lr_score_truth, 1, 0)
        lr_p = lr_p + p

    return lr_p, lr_score_truth


def lr_score_with_cluster(
        adata: anndata.AnnData,
        lr_pairs: lr_pairs,
        clusters: list,
        celltype: list,
        scale: float,
        method: str,
        radius: Union[int, float] = None,
        max_quantile: Union[int, float] = None,
        k: Union[int, float] = 1,
        m: Union[int, float] = 1,
        key: str = None,
        agg_method: str = 'gmean',
        norm: bool = True,
        percent_of_drop: Union[int, float] = 0,
        spatial_in_obsm: str = 'spatial',
        n_iters: int = 1000,
        func: bool = True,
) -> [np.ndarray, np.ndarray, list]:
    if func:
        mode = 'contact'
    else:
        mode = 'secretory'

    # Get spatial information
    indices, distances, lr_names, lr_used, exp, dict_gene = _lr_helper(adata=adata,
                                                                       lr_pairs=lr_pairs,
                                                                       scale=scale,
                                                                       key=key,
                                                                       method=method,
                                                                       radius=radius,
                                                                       max_quantile=max_quantile,
                                                                       agg_method=agg_method,
                                                                       norm=norm,
                                                                       percent_of_drop=percent_of_drop,
                                                                       spatial_in_obsm=spatial_in_obsm,
                                                                       )

    without_neighbor = 0
    neigh_num = 0
    i = 0
    for i, j in enumerate(indices):
        neigh_num += len(indices[i])

    print(f'In {mode} mode, The average number of neighbors is {neigh_num/i}')
    print(f'In {mode} mode, total of {without_neighbor} spots have no neighbors')

    for index, name in enumerate(celltype):
        index = np.int32(index)
        clusters = [index if i == name else i for i in clusters]
    clusters = np.array(clusters, dtype=np.int32)

    n_ct = np.zeros(shape=len(celltype), dtype=np.int32)
    for index_i in range(len(celltype)):
        n_ct[index_i] = np.sum(clusters == index_i)

    pd_edges = _preprocessing_of_graph(
        clu_value=clusters,
        indices=indices,
        distances=distances
    )
    count_edges = _count_edge(edge=pd_edges,
                              species_of_clusters=len(celltype),
                              cell_type_dict={ct: ct for ct in range(len(celltype))})

    count_diff_edges = np.sum(count_edges, axis=0) - count_edges.diagonal()
    expectation_edges = allocation_edge_2_diff_cell_type(count_diff_edges)

    edges = np.zeros(shape=(pd_edges.shape[0], 2))
    edges[:, 0] = pd_edges['point_1']
    edges[:, 1] = pd_edges['point_2']
    edges = edges.astype(np.int32)
    distances = pd_edges['distance'].values

    lr_names_new = []
    lr_used_new = []
    for index, name in enumerate(lr_names):
        gene_in_lr = re.split(':|&', name)
        if all(gene in adata.var_names for gene in gene_in_lr):
            lr_names_new.append(name)
            lr_used_new.append(lr_used[index])

    lr_ct_strength = np.zeros(shape=(len(lr_names_new), len(celltype), len(celltype)), dtype=np.float32)
    lr_ct_affinity = np.ones(shape=(len(lr_names_new), len(celltype), len(celltype)), dtype=np.float32) * (n_iters+1)

    with tqdm(
            total=len(lr_names_new),
            desc=f"{len(lr_names_new)} {mode} ligand-receptor pairs.",
            bar_format="{l_bar}{bar} [ time left: {remaining} ]",
    ) as pbar:

        for index, name in enumerate(lr_names_new):
            l, r = lr_used_new[index]

            lr_ct_affinity[index, :, :], lr_ct_strength[index, :, :] = get_celltype_score(
                exp_ligand=np.copy(exp[:, dict_gene[l]]),
                exp_receptor=np.copy(exp[:, dict_gene[r]]),
                len_celltype=len(celltype),
                clusters=clusters,
                edges=edges,
                distances=distances,
                func=func,
                n_iters=n_iters
            )

            l_all_mean = np.mean(exp[:, dict_gene[l]])
            r_all_mean = np.mean(exp[:, dict_gene[r]])

            for index_i in range(len(celltype)):
                for index_j in range(len(celltype)):
                    l_ct1 = np.mean(exp[np.where(clusters == index_i), dict_gene[l]])
                    r_ct2 = np.mean(exp[np.where(clusters == index_j), dict_gene[r]])

                    exp_score = np.power((l_ct1 / l_all_mean) * (r_ct2 / r_all_mean), k)
                    exp_score = np.nan_to_num(exp_score)
                    if index_i == index_j:
                        lr_ct_strength[index, index_i, index_j] = exp_score
                    else:
                        count_edge_ij = count_edges[index_i, index_j]
                        count_score = np.power(count_edge_ij/expectation_edges[index_i, index_j], m)
                        count_score = 2*count_score / (count_score + np.float32(1.0))
                        lr_ct_strength[index, index_i, index_j] = exp_score*count_score

            pbar.update(1)

    lr_ct_affinity = lr_ct_affinity / n_iters

    return [lr_ct_strength, lr_ct_affinity, lr_names_new]


def select_communication_in_celltype(
        affinity: np.ndarray,
        strength: np.ndarray,
        affinity_cutoff: float,
        strength_cutoff: float,
        celltype: list
) -> pd.DataFrame:
    """
    Count the cell-cell communication between cell types with affinity(p) and strength(s).
    """
    if affinity_cutoff is not None:
        communication_intensity_a = np.where(affinity < affinity_cutoff, 1, 0)
    else:
        communication_intensity_a = np.ones_like(affinity)

    if strength_cutoff is not None:
        communication_intensity_s = np.where(strength > strength_cutoff, 1, 0)
    else:
        communication_intensity_s = np.ones_like(strength)

    communication_intensity = communication_intensity_s & communication_intensity_a
    communication_intensity = np.sum(communication_intensity, axis=0)
    communication_intensity = pd.DataFrame(communication_intensity, index=celltype, columns=celltype)
    return communication_intensity


def cell_type_level_communication(
        adata: AnnData,
        lr_pairs: lr_pairs,
        cluster_key: str = 'cluster',
        agg_method: str = 'gmean',
        scale: Union[str, float] = 'hires',
        sample: Union[str, int, list, None] = None,
        sample_key: Optional[str] = None,
        affinity_cutoff: Optional[float] = 0.05,
        strength_cutoff: Optional[float] = 2.0,
        contact_key: Optional[str] = 'contact',
        contact_max_quantile: int = 95,
        secretory_key: Optional[str] = 'secretory',
        secretory_radius: Union[int, float] = 100,
        k: Union[str, float] = 1,
        m: Union[str, float] = 1,
        norm: bool = True,
        percent_of_drop: Union[int, float] = 0,
        spatial_in_obsm: str = 'spatial',
        n_iters: int = 1000,
        inplace: bool = True,
) -> anndata.AnnData:
    """
    Cell type ligand-receptor algorithm composed of two indexes: affinity and strength.
    The affinity is calculated from the p-value of the permutation-test.
    The strength is a combination score of the ratio of the expression and the number of edges to the expectation.

    Parameters
    ----------
    adata : anndata.Anndata
        An AnnData object containing spatial omics data and spatial information.
    lr_pairs : lr_pairs
        database of ligand-receptor pairs
    cluster_key : str
        The label of cluster in adata.obs.
    agg_method : str
        Integrated subunit to calculate ligand or receptor.
        'min': ligand or receptor = min(sub1, sub2, sub3, ...)
        'mean': ligand or receptor = mean(sub1, sub2, sub3, ...)
        'gmean': geometrical mean, ligand or receptor = np.gmean(sub1, sub2, sub3, ...)
    scale : Union[str, float]
        scale used in subsequent analyses. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this.
    sample : Union[str, int, list], optional
        Samples for which communication scores need to be calculated.
    sample_key : str, optional
        The keyword of sample id in adata.obs.columns.
    affinity_cutoff : float
        The cutoff of affinity.
    strength_cutoff : Optional[float]
        The cutoff of strength.
    contact_key : str, optional
        The tag name that represents the contact type in the LR database
    contact_max_quantile : int
        In Neighbor network, Order the distance of all sides, and more than max_quantile% of the sides will be removed
    secretory_key : str, optional
        The tag name that represents the secretory type in the LR database
    secretory_radius : Union[int, float]
        The maximum distance considered for the secretory type
    k: Union[str, float]
        The weight of the expression function
    m: Union[str, float]
        The weight of the edge function
    norm : bool
        Whether 0-1 normalization was performed on the expression volume data.
    percent_of_drop : Union[int, float]
        Percentage of extreme values removed from normalization
    spatial_in_obsm : str
        The key of spatial coordinates in adata.obsm
    n_iters : int
        Number of permutations for the permutation-test
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    - :attr:`anndata.AnnData.uns` ``['SOAPy']['ligand_spot_comm_score' or 'receptor_spot_comm_score']``
        - 'celltype': cell type in used sample.
        - 'contact': cell-cell communication between cell types under contact type LR.
            - 'names': Names of used contact type lR pairs.
            - 'sig_celltype': How cell types communicate with each other under contact conditions.
            - 'strength': The strength of contact type lR pairs.
            - 'affinity': The affinity of contact type lR pairs.
        - 'secretory': cell-cell communication between cell types under secretory type LR.
            - 'names': Names of used secretory type lR pairs.
            - 'sig_celltype': How cell types communicate with each other under secretory conditions.
            - 'strength': The strength of secretory type lR pairs.
            - 'affinity': The affinity of secretory type lR pairs.

    """
    adata = _check_adata_type(adata, spatial_in_obsm, inplace)

    if isinstance(sample, str):
        sample = [sample]
    if sample is not None and sample_key is None:
        logg.error(f'Mult-sample niche analysis cannot be specified without a given sample key')
        raise ValueError
    if sample is None and sample_key is not None:
        sample = list(np.unique(adata.obs[sample_key]))
        logg.info(f'Use all samples in the niche calculation')
    if sample is None and sample_key is None:
        sample = [None]

    for sample_id in sample:
        if sample_id is None:
            bdata = adata
        else:
            bdata = adata[adata.obs[sample_key] == sample, :].copy()

        if type(scale) != float:
            scale = _scale(bdata, None, scale)

        clusters = adata.obs[cluster_key].tolist()
        celltype = list(np.unique(clusters))

        if contact_key is None:
            c_sig_celltype = None
            c_name = None
            contact_strength_ct = None
            contact_affinity_ct = None
        else:
            contact_strength_ct, contact_affinity_ct, c_name = lr_score_with_cluster(
                adata=bdata,
                clusters=clusters,
                celltype=celltype,
                scale=scale,
                method='neighbor',
                k=k,
                m=m,
                max_quantile=contact_max_quantile,
                lr_pairs=lr_pairs,
                key=contact_key,
                agg_method=agg_method,
                norm=norm,
                percent_of_drop=percent_of_drop,
                spatial_in_obsm=spatial_in_obsm,
                n_iters=n_iters,
                func=True
            )

            c_sig_celltype = select_communication_in_celltype(affinity=contact_affinity_ct,
                                                              strength=contact_strength_ct,
                                                              affinity_cutoff=affinity_cutoff,
                                                              strength_cutoff=strength_cutoff,
                                                              celltype=list(celltype))

        if secretory_key is None:
            s_sig_celltype = None
            s_name = None
            secretory_strength_ct = None
            secretory_affinity_ct = None
        else:
            secretory_strength_ct, secretory_affinity_ct, s_name = lr_score_with_cluster(
                adata=bdata,
                clusters=clusters,
                celltype=celltype,
                scale=scale,
                method='radius',
                radius=secretory_radius,
                k=k,
                m=m,
                lr_pairs=lr_pairs,
                key=secretory_key,
                agg_method=agg_method,
                norm=norm,
                percent_of_drop=percent_of_drop,
                spatial_in_obsm=spatial_in_obsm,
                n_iters=n_iters,
                func=False
            )

            s_sig_celltype = select_communication_in_celltype(affinity=secretory_affinity_ct,
                                                              strength=secretory_strength_ct,
                                                              affinity_cutoff=affinity_cutoff,
                                                              strength_cutoff=strength_cutoff,
                                                              celltype=list(celltype))

        communication_ = {
            'celltype': celltype,
            'contact': {
                'names': c_name,
                'sig_celltype': c_sig_celltype,
                'strength': contact_strength_ct,
                'affinity': contact_affinity_ct
            },
            'secretory': {
                'names': s_name,
                'sig_celltype': s_sig_celltype,
                'strength': secretory_strength_ct,
                'affinity': secretory_affinity_ct
            }
        }
        _add_info_from_sample(adata, sample_id=sample_id, keys='celltype_comm_score', add=communication_)

    return adata



import copy
import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as st
from typing import Any, Tuple, Optional, Union, Sequence, Literal
from matplotlib.colors import ListedColormap
from os import PathLike
from matplotlib.axes import Axes
from ..utils import _get_info_from_sample
from ._color import _get_palette, color_list_50

__all__ = [
    'show_tendency',
    'show_curves_cluster',
    'show_box_plot'
]


def show_tendency(
        adata: ad.AnnData,
        gene_name: Union[str, list],
        method: str = 'poly',
        one_axes: bool = True,
        palette: Union[Sequence[str], ListedColormap, dict] = None,
        norm: bool = False,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: int = 100,
        grid: bool = False,
        show: bool = True,
        save: Union[str, PathLike, None] = None,
        legend_kwargs: dict = {},
        **kwargs: Any
) -> Union[Axes, Sequence[Axes], None]:
    """
    Plot the tendency curve of the genes.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    gene_name : Union[list, str, None], optional
        The gene names for the regression model need to be calculated.
    method : Literal['poly', 'loess'], optional
        Polynomial regression(poly) or Loess regression(loess).
    one_axes : bool
        Whether to plot all curves on the same axis.
    palette : Union[Sequence[str], Cycler, dict], optional
        Colors to use for plotting annotation groups.
    norm : bool
        Normalize the values of the curves.
    figsize : Tuple[Any, Any]
        The size of format figure.(width, height).
    dpi : int
        Dots per inch values for the output.
    grid : bool
        Whether to show the grid or not.
    show : bool
        Show this plot.
    save : Union[str, PathLike]
        The path where the image is stored.
    legend_kwargs : dict, optional
        Other params of plt.legend(). Only used in 'one_axes == True'.
    kwargs
        Other params of ax.plot().

    Returns
    -------
        If show==False, return Axes or a list of Axes.
    """
    adata = copy.deepcopy(adata)

    if type(gene_name) == str:
        gene_name = [gene_name]

    if method == 'poly':
        param = _get_info_from_sample(adata, sample_id=None, key='poly')
        dic_crd = param['dic_crd_poly']

    elif method == 'loess':
        param = _get_info_from_sample(adata, sample_id=None, key='loess')
        dic_crd = param['dic_crd_loess']
    else:
        raise ValueError

    x = dic_crd['Xest']
    dic_crd.pop('Xest')
    axes = []####

    if not isinstance(palette, dict):
        palette = _get_palette(gene_name, sort_order=False, palette=palette)

    if norm:
        for gene in gene_name:
            y = dic_crd[gene]
            ygrid = []
            ygrid = (np.array(ygrid) - min(y)) / (max(y) - min(y))
            dic_crd[gene] = ygrid

    if one_axes:
        if figsize is None:
            figsize = (8, 6)

        fig = plt.figure(figsize=figsize,
                         dpi=dpi,
                         )

        ax = fig.add_axes([0.1, 0.1, 0.7, 0.75])
        for i, gene in enumerate(gene_name):
            color = palette[gene]

            ax.plot(x, dic_crd[gene],
                    color=color,
                    label=gene,
                    **kwargs
                    )

            if 'bbox_to_anchor' not in legend_kwargs.keys():
                legend_kwargs['bbox_to_anchor'] = (1.02, 0.8)
            if 'loc' not in legend_kwargs.keys():
                legend_kwargs['loc'] = 'upper left'
            if 'title' not in legend_kwargs.keys():
                legend_kwargs['title'] = 'gene'

            ax.legend(**legend_kwargs)

        axes = ax

        if grid:
            ax.grid()
        ax.set(xlabel='distance',
               ylabel='expression',
               title=method
               )

    else:
        length = len(gene_name)

        if figsize is None:
            figsize = (8 * length, 6)

        fig = plt.figure(figsize=figsize,
                         dpi=dpi,
                         )

        gap = 0
        for i, gene in enumerate(gene_name):
            if i == 0:
                ax = fig.add_axes([0.05, 0.15, 0.8 / length, 0.7])
                gap += 0.05 + 0.8 / length
            else:
                ax = fig.add_axes([gap + 0.1/length, 0.15, 0.8 / length, 0.7])
                gap += 0.9/length

            color = palette[gene]

            ax.plot(x, dic_crd[gene],
                    color=color,
                    **kwargs
                    )

            ax.set(xlabel='distance',
                   ylabel='expression',
                   title=gene)
            if grid:
                ax.grid()
            axes.append(ax)
    plt.tight_layout()

    if save:
        fig.savefig(save)
    if show:
        fig.show()
    else:
        return axes


def show_curves_cluster(
        adata: ad.AnnData,
        method: Literal['poly', 'loess'] = 'poly',
        show_standard_error: bool = True,
        SE_alpha: Optional[float] = None,
        palette: Union[Sequence[str], ListedColormap] = None,
        title: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: int = 100,
        ax: Optional[Axes] = None,
        grid: bool = False,
        show: bool = True,
        save: Union[str, PathLike, None] = None,
        legend_kwargs: dict = {},
        **kwargs: Any
) -> Optional[Axes]:
    """
    Plot the tendency curve of the gene clusters.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    ax : Axes
        A matplotlib axes object.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    method : Literal['poly', 'loess'], optional
        Polynomial regression(poly) or Loess regression(loess).
    show_standard_error : bool
        Plot the standard error margin of the clustering curve.
    SE_alpha : bool
        Transparency between the drawn standard errors
    palette : Union[Sequence[str], ListedColormap], optional
        Colors to use for plotting annotation groups.
    title : str, optional
        The title of figure
    grid : bool
        Whether to add grid or not.
    legend_kwargs : dict, optional
        Other params of plt.legend(). Only used in 'one_axes == True'.
    kwargs : Any
        Other params of ax.plot()

    Returns
    -------
        If show==False, return Axes.
    """
    adata = copy.deepcopy(adata)

    gene_params = _get_info_from_sample(adata, sample_id=None, key=method)
    dic_crd = gene_params['dic_crd_' + method]
    param = gene_params['df_param_' + method]
    cluster_params = _get_info_from_sample(adata, sample_id=None, key='gene_cluster')
    k = cluster_params['k']
    gene_cluster = cluster_params['gene_cluster']

    if figsize is None:
        figsize = (10, 8)

    if title is None:
        title = 'Gene cluster'

    if palette is None:
        palette = color_list_50
    if isinstance(palette, ListedColormap):
        palette = palette.colors

    if ax is None:
        fig = plt.figure(figsize=figsize,
                         dpi=dpi,
                         )

        ax = fig.add_axes([0.15, 0.15, 0.7, 0.7])

    for i in range(k):

        data_new = gene_cluster[gene_cluster['cluster'] == i]
        gene_name = data_new.index

        x = dic_crd['Xest']
        x_num = len(x)
        matrix_y = []
        if method == 'poly':
            xgrid = np.linspace(min(x), max(x), num=x_num)
            for gene in gene_name:
                y = dic_crd[gene]
                params = param.loc[gene, 'param']
                ygrid = []
                for x0 in xgrid:
                    ygrid.append(sum([param * x0 ** i for i, param in enumerate(params)]))
                ygrid = (np.array(ygrid) - min(y)) / (max(y) - min(y))
                matrix_y.append(ygrid)
        else:
            samples = np.random.choice(len(x), x_num, replace=True)
            samples = np.sort(samples)
            xgrid = x[samples]
            for gene in gene_name:
                y = dic_crd[gene]
                ygrid = y[samples]
                ygrid = (np.array(ygrid) - min(y)) / (max(y) - min(y))
                matrix_y.append(ygrid)

        matrix_y = np.array(matrix_y)
        meta_y = np.mean(matrix_y, 0)
        #meta_y = (meta_y - meta_y.min()) / (meta_y.max() - meta_y.min())

        ax.plot(xgrid, meta_y, c=palette[i], label='cluster ' + str(i), **kwargs)

        if show_standard_error and matrix_y.shape[0] > 1:
            if SE_alpha is None:
                SE_alpha = 0.3
            low_CI_bound, high_CI_bound = st.t.interval(0.95, matrix_y.shape[0] - 1,
                                                        loc=meta_y,
                                                        scale=st.sem(matrix_y))

            ax.fill_between(xgrid, low_CI_bound, high_CI_bound, alpha=SE_alpha, color=palette[i])

    ax.set(title=title, xlabel='distance', ylabel='norm expression')

    if 'bbox_to_anchor' not in legend_kwargs.keys():
        legend_kwargs['bbox_to_anchor'] = (1.02, 0.8)
    if 'loc' not in legend_kwargs.keys():
        legend_kwargs['loc'] = 'upper left'
    if 'title' not in legend_kwargs.keys():
        legend_kwargs['title'] = 'gene cluster'
    ax.legend(**legend_kwargs)

    plt.tight_layout()

    if grid:
        ax.grid()
    if save:
        plt.savefig(save)
    if show:
        plt.show()
    else:
        return ax


def show_box_plot(
        adata: ad.AnnData,
        cluster_key: str,
        score_key: str,
        title: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = None,
        dpi: int = 100,
        ax: Optional[Axes] = None,
        show: bool = True,
        save: Union[str, PathLike, None] = None,
        **kwargs: Any,
) -> Optional[Axes]:
    """
    Boxplot of gene expression for each cluster is shown

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    ax : Axes
        A matplotlib axes object.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    cluster_key
        cluster keyword in adata.obs.index.
    score_key
        Gene name in adata.var_names.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike]
        The path where the image is stored.
    kwargs
        Other params of ax.plot().

    Returns
    -------

    """
    import pandas as pd
    import seaborn as sns

    if title is None:
        title = 'ANOVA'

    cluster = adata.obs[cluster_key]
    if score_key in adata.obs.columns:
        score = np.array(adata.obs[score_key])
    elif score_key in adata.var_names:
        index = adata.var_names.tolist().index(score_key)
        score = adata.X.toarray()[:, index]
    else:
        raise ValueError

    df = pd.DataFrame({'cluster': cluster, 'score': score})

    if ax is None:
        if figsize is None:
            figsize = (8, 6)

        fig = plt.figure(figsize=figsize, dpi=dpi)
        ax = fig.add_subplot(111)

    ax = sns.boxplot(x='cluster', y='score', ax=ax, data=df, **kwargs)

    ax.set(title=title, xlabel=cluster_key, ylabel=score_key)
    if save:
        plt.savefig(save)
    if show:
        plt.show()
    else:
        return ax


from ._show import show_moran_scatterplot, show_network, show_voronoi
from ._tendency import show_tendency, show_curves_cluster, show_box_plot
from ._neighborhood import show_neighborhood_analysis, show_infiltration_analysis
from ._ccc import show_ccc_netplot, show_ccc_chordplot, show_ccc_dotplot, show_ccc_embedding
from ._niche import show_niche_environment, show_niche_sample_heatmap, show_celltype_niche_heatmap, \
    show_celltype_sample_heatmap
from ._tensor import show_factor_matrix_in_CP_tensor, show_factor_matrix_in_tucker_tensor, show_proportion_in_tucker, \
    show_proportion_in_CP



import warnings
import copy
import anndata
import numpy as np
import pandas as pd
from typing import Optional, Union, Tuple, Literal
from .utils import _filter_of_graph, _preprocessing_of_graph
from ..utils import _scale, _graph, _add_info_from_sample, _check_adata_type

__all__ = ["make_network"]


def make_network(
        adata: anndata.AnnData,
        sample_key: Optional[str] = None,
        sample: Union[str, int, list, None] = None,
        cluster_key: str = 'clusters',
        method: Literal['radius', 'knn', 'regular', 'neighbor'] = 'knn',
        cutoff: Union[float, int] = 6,
        max_quantile: Union[float, int] = 98,
        exclude: Union[str, dict] = None,
        scale: Union[str, float] = 'hires',
        spatial_in_obsm: str = 'spatial',
        inplace: bool = True
) -> anndata.AnnData:
    """
    A function to create a network based on spatial information.
    We offer four different ways to build a network: KNN network, range radiation network, regular network and First-order neighbor network

    'exclude' is a parameter to exclude categories that cannot form an edge, you can set 'same' and 'different' to
    specifies the same/different clusters may not be connected.If you want to define a custom class of points that
    can't be connected as edges, pass it as a dictionary.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    sample_key : str, optional
        Batch's key in adata.obs.
    sample : Union[str, int, list], optional
        The sample number for which the network needs to be built.
    cluster_key : str, optional
        The column label of clusters in adata.obs.
    method : str, optional
        the method to make network, select in 'Radius' and 'KNN'.
    cutoff : Union[float, int], optional
        In KNN network and regular network, cutoff means number of neighbors to use. In range radiation network, cutoff means range of parameter space to use
    max_quantile : Union[float, int], optional
        In Neighbor network, Order the distance of all sides, and more than max_quantile% of the sides will be removed
    exclude : Union[str, float], optional
        Excluding categories that cannot form an edge.
    scale : Union[str, float], optional
        Scale used in subsequent analyses. If it's Visium data it can also be HE image labels (hires or lower).
        Most of the time you don't need to change this.
    spatial_in_obsm : str, optional
        Keyword of coordinate information in obsm.
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    :attr:`~anndata.AnnData.uns['SOAPy']`
        SOAPy generated parameters
    :attr:`~anndata.AnnData.uns['SOAPy']['indices']`
        adjacency matrix of network
    :attr:`~anndata.AnnData.uns['SOAPy']['distance']`
        distance matrix of network
    :attr:`~anndata.AnnData.uns['SOAPy']['edges']`
        edges of network

    """
    adata = _check_adata_type(adata, spatial_in_obsm, inplace)

    df_pixel = adata.obsm[spatial_in_obsm]
    df_pixel = pd.DataFrame(df_pixel)
    df_pixel = df_pixel.iloc[:, [0, 1]]

    if sample_key is None:
        if type(scale) != float:
            scale = _scale(adata, None, scale)

        indices, distances = _graph(
            col=list(df_pixel.iloc[:, 0] * scale),
            row=list(df_pixel.iloc[:, 1] * scale),
            method=method,
            cutoff=cutoff,
            max_quantile=max_quantile
        )

        _add_info_from_sample(adata,
                              keys=['indices', 'distance', 'edges'],
                              add=_get_edge(adata.obs, indices, distances, cluster_key, exclude))
    else:
        df_pixel['sample'] = adata.obs[sample_key].tolist()

        if sample is None:
            sample = df_pixel['sample'].unique().tolist()
        if not isinstance(sample, list):
            sample = [sample]
        for index in sample:
            df_pixel_sample = df_pixel[df_pixel['sample'] == index]
            obs = adata.obs[adata.obs[sample_key] == index]

            if type(scale) != float:
                scale = _scale(adata, index, scale)

            indices, distances = _graph(
                col=df_pixel_sample.iloc[:, 0] * scale,
                row=df_pixel_sample.iloc[:, 1] * scale,
                method=method,
                cutoff=cutoff,
                max_quantile=max_quantile
            )

            _add_info_from_sample(adata,
                                  sample_id=index,
                                  keys=['indices', 'distance', 'edges'],
                                  add=_get_edge(obs, indices, distances, cluster_key, exclude))

    return adata


def _get_edge(obs, indices, distances, cluster_label, exclude):
    """
    By adding constraints to the network, the modified advanced matrix and edge are obtained
    """
    if exclude is not None:
        indices, distances = _filter_of_graph(
            adata=obs,
            indices=indices,
            distances=distances,
            cluster_label=cluster_label,
            exclude=exclude
        )

    edge = _preprocessing_of_graph(
        clu_value=obs[cluster_label].values,
        indices=indices,
        distances=distances
    )
    return indices, distances, edge


from ._mask import get_mask_from_domain
from ._domain import domain_from_unsupervised, global_moran, cal_aucell, domain_from_local_moran
from ._var_genes import cal_spatialDE, cal_sparkX, cal_spark
from ._tendency import adata_from_mask, wilcoxon_test, spearman_correlation, ANOVA, spatial_tendency, gene_cluster
from ._interaction import neighborhood_analysis, infiltration_analysis, get_c_niche
from ._ccc import cell_level_communications, cell_type_level_communication, lr_pairs
from ._tensor import TensorDecomposition



import torch
from .preprocess import preprocess_adj, preprocess_adj_sparse, preprocess, construct_interaction, construct_interaction_KNN, add_contrastive_label, get_feature, permutation, fix_seed
import time
import random
import numpy as np
from .model import Encoder, Encoder_sparse, Encoder_map, Encoder_sc
from tqdm import tqdm
from torch import nn
import torch.nn.functional as F
from scipy.sparse.csc import csc_matrix
from scipy.sparse.csr import csr_matrix
import pandas as pd
    
class GraphST():
    def __init__(self, 
        adata,
        adata_sc = None,
        device= torch.device('cpu'),
        learning_rate=0.001,
        learning_rate_sc = 0.01,
        weight_decay=0.00,
        epochs=600, 
        dim_input=3000,
        dim_output=64,
        random_seed = 41,
        alpha = 10,
        beta = 1,
        theta = 0.1,
        lamda1 = 10,
        lamda2 = 1,
        deconvolution = False,
        datatype = '10X'
        ):
        '''\

        Parameters
        ----------
        adata : anndata
            AnnData object of spatial data.
        adata_sc : anndata, optional
            AnnData object of scRNA-seq data. adata_sc is needed for deconvolution. The default is None.
        device : string, optional
            Using GPU or CPU? The default is 'cpu'.
        learning_rate : float, optional
            Learning rate for ST representation learning. The default is 0.001.
        learning_rate_sc : float, optional
            Learning rate for scRNA representation learning. The default is 0.01.
        weight_decay : float, optional
            Weight factor to control the influence of weight parameters. The default is 0.00.
        epochs : int, optional
            Epoch for model training. The default is 600.
        dim_input : int, optional
            Dimension of input feature. The default is 3000.
        dim_output : int, optional
            Dimension of output representation. The default is 64.
        random_seed : int, optional
            Random seed to fix model initialization. The default is 41.
        alpha : float, optional
            Weight factor to control the influence of reconstruction loss in representation learning. 
            The default is 10.
        beta : float, optional
            Weight factor to control the influence of contrastive loss in representation learning. 
            The default is 1.
        lamda1 : float, optional
            Weight factor to control the influence of reconstruction loss in mapping matrix learning. 
            The default is 10.
        lamda2 : float, optional
            Weight factor to control the influence of contrastive loss in mapping matrix learning. 
            The default is 1.
        deconvolution : bool, optional
            Deconvolution task? The default is False.
        datatype : string, optional    
            Data type of input. Our model supports 10X Visium ('10X'), Stereo-seq ('Stereo'), and Slide-seq/Slide-seqV2 ('Slide') data. 
        Returns
        -------
        The learned representation 'self.emb_rec'.

        '''
        self.adata = adata.copy()
        self.device = device
        self.learning_rate=learning_rate
        self.learning_rate_sc = learning_rate_sc
        self.weight_decay=weight_decay
        self.epochs=epochs
        self.random_seed = random_seed
        self.alpha = alpha
        self.beta = beta
        self.theta = theta
        self.lamda1 = lamda1
        self.lamda2 = lamda2
        self.deconvolution = deconvolution
        self.datatype = datatype
        
        fix_seed(self.random_seed)
        
        if 'highly_variable' not in adata.var.keys():
           preprocess(self.adata)
        
        if 'adj' not in adata.obsm.keys():
           if self.datatype in ['Stereo', 'Slide']:
              construct_interaction_KNN(self.adata)
           else:    
              construct_interaction(self.adata)
         
        if 'label_CSL' not in adata.obsm.keys():    
           add_contrastive_label(self.adata)
           
        if 'feat' not in adata.obsm.keys():
           get_feature(self.adata)
        
        self.features = torch.FloatTensor(self.adata.obsm['feat'].copy()).to(self.device)
        self.features_a = torch.FloatTensor(self.adata.obsm['feat_a'].copy()).to(self.device)
        self.label_CSL = torch.FloatTensor(self.adata.obsm['label_CSL']).to(self.device)
        self.adj = self.adata.obsm['adj']
        self.graph_neigh = torch.FloatTensor(self.adata.obsm['graph_neigh'].copy() + np.eye(self.adj.shape[0])).to(self.device)
    
        self.dim_input = self.features.shape[1]
        self.dim_output = dim_output
        
        if self.datatype in ['Stereo', 'Slide']:
           #using sparse
           print('Building sparse matrix ...')
           self.adj = preprocess_adj_sparse(self.adj).to(self.device)
        else: 
           # standard version
           self.adj = preprocess_adj(self.adj)
           self.adj = torch.FloatTensor(self.adj).to(self.device)
        
        if self.deconvolution:
           self.adata_sc = adata_sc.copy() 
            
           if isinstance(self.adata.X, csc_matrix) or isinstance(self.adata.X, csr_matrix):
              self.feat_sp = adata.X.toarray()[:, ]
           else:
              self.feat_sp = adata.X[:, ]
           if isinstance(self.adata_sc.X, csc_matrix) or isinstance(self.adata_sc.X, csr_matrix):
              self.feat_sc = self.adata_sc.X.toarray()[:, ]
           else:
              self.feat_sc = self.adata_sc.X[:, ]
            
           # fill nan as 0
           self.feat_sc = pd.DataFrame(self.feat_sc).fillna(0).values
           self.feat_sp = pd.DataFrame(self.feat_sp).fillna(0).values
          
           self.feat_sc = torch.FloatTensor(self.feat_sc).to(self.device)
           self.feat_sp = torch.FloatTensor(self.feat_sp).to(self.device)
        
           if self.adata_sc is not None:
              self.dim_input = self.feat_sc.shape[1] 

           self.n_cell = adata_sc.n_obs
           self.n_spot = adata.n_obs
            
    def train(self):
        if self.datatype in ['Stereo', 'Slide']:
           self.model = Encoder_sparse(self.dim_input, self.dim_output, self.graph_neigh).to(self.device)
        else:
           self.model = Encoder(self.dim_input, self.dim_output, self.graph_neigh).to(self.device)
        self.loss_CSL = nn.BCEWithLogitsLoss()
    
        self.optimizer = torch.optim.Adam(self.model.parameters(), self.learning_rate, 
                                          weight_decay=self.weight_decay)
        
        print('Begin to train ST data...')
        self.model.train()
        
        for epoch in tqdm(range(self.epochs)): 
            self.model.train()
              
            self.features_a = permutation(self.features)
            self.hiden_feat, self.emb, ret, ret_a = self.model(self.features, self.features_a, self.adj)
            
            self.loss_sl_1 = self.loss_CSL(ret, self.label_CSL)
            self.loss_sl_2 = self.loss_CSL(ret_a, self.label_CSL)
            self.loss_feat = F.mse_loss(self.features, self.emb)
            
            loss =  self.alpha*self.loss_feat + self.beta*(self.loss_sl_1 + self.loss_sl_2)
            
            self.optimizer.zero_grad()
            loss.backward() 
            self.optimizer.step()
        
        print("Optimization finished for ST data!")
        
        with torch.no_grad():
             self.model.eval()
             if self.deconvolution:
                self.emb_rec = self.model(self.features, self.features_a, self.adj)[1]
                
                return self.emb_rec
             else:  
                if self.datatype in ['Stereo', 'Slide']:
                   self.emb_rec = self.model(self.features, self.features_a, self.adj)[1]
                   self.emb_rec = F.normalize(self.emb_rec, p=2, dim=1).detach().cpu().numpy() 
                else:
                   self.emb_rec = self.model(self.features, self.features_a, self.adj)[1].detach().cpu().numpy()
                self.adata.obsm['emb'] = self.emb_rec
                
                return self.adata
         
    def train_sc(self):
        self.model_sc = Encoder_sc(self.dim_input, self.dim_output).to(self.device)
        self.optimizer_sc = torch.optim.Adam(self.model_sc.parameters(), lr=self.learning_rate_sc)  
        
        print('Begin to train scRNA data...')
        for epoch in tqdm(range(self.epochs)):
            self.model_sc.train()
            
            emb = self.model_sc(self.feat_sc)
            loss = F.mse_loss(emb, self.feat_sc)
            
            self.optimizer_sc.zero_grad()
            loss.backward()
            self.optimizer_sc.step()
            
        print("Optimization finished for cell representation learning!")
        
        with torch.no_grad():
            self.model_sc.eval()
            emb_sc = self.model_sc(self.feat_sc)
         
            return emb_sc
        
    def train_map(self):
        emb_sp = self.train()
        emb_sc = self.train_sc()
        
        self.adata.obsm['emb_sp'] = emb_sp.detach().cpu().numpy()
        self.adata_sc.obsm['emb_sc'] = emb_sc.detach().cpu().numpy()
        
        # Normalize features for consistence between ST and scRNA-seq
        emb_sp = F.normalize(emb_sp, p=2, eps=1e-12, dim=1)
        emb_sc = F.normalize(emb_sc, p=2, eps=1e-12, dim=1)
        
        self.model_map = Encoder_map(self.n_cell, self.n_spot).to(self.device)  
          
        self.optimizer_map = torch.optim.Adam(self.model_map.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay)
        
        print('Begin to learn mapping matrix...')
        for epoch in tqdm(range(self.epochs)):
            self.model_map.train()
            self.map_matrix = self.model_map()

            loss_recon, loss_NCE = self.loss(emb_sp, emb_sc)
             
            loss = self.lamda1*loss_recon + self.lamda2*loss_NCE 

            self.optimizer_map.zero_grad()
            loss.backward()
            self.optimizer_map.step()
            
        print("Mapping matrix learning finished!")
        
        # take final softmax w/o computing gradients
        with torch.no_grad():
            self.model_map.eval()
            emb_sp = emb_sp.cpu().numpy()
            emb_sc = emb_sc.cpu().numpy()
            map_matrix = F.softmax(self.map_matrix, dim=1).cpu().numpy() # dim=1: normalization by cell
            
            self.adata.obsm['emb_sp'] = emb_sp
            self.adata_sc.obsm['emb_sc'] = emb_sc
            self.adata.obsm['map_matrix'] = map_matrix.T # spot x cell

            return self.adata, self.adata_sc
    
    def loss(self, emb_sp, emb_sc):
        '''\
        Calculate loss

        Parameters
        ----------
        emb_sp : torch tensor
            Spatial spot representation matrix.
        emb_sc : torch tensor
            scRNA cell representation matrix.

        Returns
        -------
        Loss values.

        '''
        # cell-to-spot
        map_probs = F.softmax(self.map_matrix, dim=1)   # dim=0: normalization by cell
        self.pred_sp = torch.matmul(map_probs.t(), emb_sc)
           
        loss_recon = F.mse_loss(self.pred_sp, emb_sp, reduction='mean')
        loss_NCE = self.Noise_Cross_Entropy(self.pred_sp, emb_sp)
           
        return loss_recon, loss_NCE
        
    def Noise_Cross_Entropy(self, pred_sp, emb_sp):
        '''\
        Calculate noise cross entropy. Considering spatial neighbors as positive pairs for each spot
            
        Parameters
        ----------
        pred_sp : torch tensor
            Predicted spatial gene expression matrix.
        emb_sp : torch tensor
            Reconstructed spatial gene expression matrix.

        Returns
        -------
        loss : float
            Loss value.

        '''
        
        mat = self.cosine_similarity(pred_sp, emb_sp) 
        k = torch.exp(mat).sum(axis=1) - torch.exp(torch.diag(mat, 0))
        
        # positive pairs
        p = torch.exp(mat)
        p = torch.mul(p, self.graph_neigh).sum(axis=1)
        
        ave = torch.div(p, k)
        loss = - torch.log(ave).mean()
        
        return loss
    
    def cosine_similarity(self, pred_sp, emb_sp):  #pres_sp: spot x gene; emb_sp: spot x gene
        '''\
        Calculate cosine similarity based on predicted and reconstructed gene expression matrix.    
        '''
        
        M = torch.matmul(pred_sp, emb_sp.T)
        Norm_c = torch.norm(pred_sp, p=2, dim=1)
        Norm_s = torch.norm(emb_sp, p=2, dim=1)
        Norm = torch.matmul(Norm_c.reshape((pred_sp.shape[0], 1)), Norm_s.reshape((emb_sp.shape[0], 1)).T) + -5e-12
        M = torch.div(M, Norm)
        
        if torch.any(torch.isnan(M)):
           M = torch.where(torch.isnan(M), torch.full_like(M, 0.4868), M)

        return M        



import os.path as osp

import argparse
import torch
import torch.nn.functional as F
from torch_geometric.datasets import Planetoid
import torch_geometric.transforms as T
from torch_geometric.nn import GCNConv, GAE, VGAE

import pickle
import numpy as np

def sparse_mx_to_torch_edge_list(sparse_mx):
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    edge_list = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    return edge_list

def rep_gae(n_h, X, A, mdl_type="VGAE"):

    n_nd, n_f = X.shape

    class Encoder(torch.nn.Module):
        def __init__(self, in_channels, out_channels):
            super(Encoder, self).__init__()
            self.conv1 = GCNConv(in_channels, 2 * out_channels, cached=True)
            if mdl_type == 'GAE':
                self.conv2 = GCNConv(2 * out_channels, out_channels, cached=True)
            elif mdl_type == 'VGAE':
                self.conv_mu = GCNConv(2 * out_channels, out_channels, cached=True)
                self.conv_logvar = GCNConv(
                    2 * out_channels, out_channels, cached=True)

        def forward(self, x, edge_index):
            x = F.relu(self.conv1(x, edge_index))
            if mdl_type == 'GAE':
                return self.conv2(x, edge_index)
            elif mdl_type == 'VGAE':
                return self.conv_mu(x, edge_index), self.conv_logvar(x, edge_index)

    channels = n_h
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    if mdl_type == 'GAE':
        model = GAE(Encoder(n_f, channels)).to(device)
    elif mdl_type == 'VGAE':
        model = VGAE(Encoder(n_f, channels)).to(device)
    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)
    X = torch.FloatTensor(X)
    X = X.to(device)
    edge_list = sparse_mx_to_torch_edge_list(A)
    edge_list = edge_list.to(device)

    def train():
        model.train()
        optimizer.zero_grad()
        z = model.encode(X, edge_list)
        loss = model.recon_loss(z, edge_list)
        if mdl_type == 'VGAE':
            loss = loss + (1 / n_nd) * model.kl_loss()
        loss.backward()
        optimizer.step()
        return loss.item()
    
    for epoch in range(1000):
        loss = train()
        print('Epoch: {:03d}, Loss: {:.4f}'.format(epoch, loss))
    
    z,_,_ = model(X, edge_list)
    return z.cpu().detach().numpy()

def main():
    f = open("./X.pkl", 'rb')
    X = pickle.load(f)
    X = X.todense()
    f = open("./A.pkl", 'rb')
    A = pickle.load(f)
    rep_gae(100, X, A, mdl_type='VGAE')

if __name__=="__main__":
    main()



import numpy as np
import numba as nb
import pandas as pd
from anndata import AnnData
from typing import Union, Optional, Tuple


@nb.jit
def _filter_of_graph(obs: pd.DataFrame,
                     indices: np.ndarray,
                     distances: np.ndarray,
                     cluster_label: str,
                     exclude: Union[str, dict]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Select the edges as 'exclude'

    Returns
    -------
        anndata.Anndata object.

    """
    if type(exclude) == str:
        assert exclude in ['same', 'different']
    if type(exclude) == dict:
        exclude = [(i, j) for i, j in zip(exclude.keys(), exclude.values())] + \
                  [(j, i) for i, j in zip(exclude.keys(), exclude.values())]

    for i, neigh in enumerate(indices):
        point_1 = neigh[0]
        index_point1 = obs.index[point_1]
        for j, point_2 in enumerate(neigh):
            if j == 0:
                continue
            index_point2 = obs.index[point_2]
            if exclude == 'same' and \
                    obs.loc[index_point1, cluster_label] == obs.loc[index_point2, cluster_label]:
                distances[i][j] = -1
                indices[i][j] = -1
                continue
            if exclude == 'different' and \
                    obs.loc[index_point1, cluster_label] != obs.loc[index_point2, cluster_label]:
                distances[i][j] = -1
                indices[i][j] = -1
                continue
            if type(exclude) == dict and \
                    (obs.loc[index_point1, cluster_label], obs.loc[index_point2, cluster_label]) in exclude:
                distances[i][j] = -1
                indices[i][j] = -1
        distances[i] = distances[i][distances[i] >= 0]
        indices[i] = indices[i][indices[i] >= 0]

    return indices, distances


def _preprocessing_of_graph(clu_value: np.ndarray,
                            indices: np.ndarray,
                            distances: np.ndarray,
                            ) -> pd.DataFrame:
    """
    get information of edges
    """
    edges = []

    for index, neigh in enumerate(indices):
        if len(neigh) == 0:
            continue
        for index_2, point_2 in enumerate(neigh):
            if point_2 == index | point_2 == -1:
                continue
            elif point_2 < index:
                edge = [point_2, index,
                        clu_value[point_2],
                        clu_value[index],
                        distances[index][index_2]]
                edges.append(edge)

            elif point_2 > index:
                edge = [index, point_2,
                        clu_value[index],
                        clu_value[point_2],
                        distances[index][index_2]]
                edges.append(edge)

    df_edge = pd.DataFrame(data=np.array(edges), columns=['point_1', 'point_2',
                                                          'cluster_1', 'cluster_2', 'distance'])

    df_edge.drop_duplicates(subset=['point_1', 'point_2'], inplace=True)
    return df_edge



#!/usr/bin/env python
"""
# Author: Yahui Long
# File Name: __init__.py
# Description:
"""

__author__ = "Yahui Long"
__email__ = "long_yahui@immunol.a-star.edu.sg"

from .utils import clustering, project_cell_to_spot
from .preprocess import preprocess_adj, preprocess, construct_interaction, add_contrastive_label, get_feature, permutation, fix_seed



import copy
import anndata
import numpy as np
import pandas as pd
from typing import Optional, Union, Literal
from ..utils import _add_info_from_sample, _get_info_from_sample, _check_adata_type
from .utils import (
    _count_edge,
    _randomize_helper,
    Iterators,
    _best_k,
)
import numba as nb
from joblib import Parallel, delayed
import logging as logg

__all__ = ['neighborhood_analysis', 'infiltration_analysis', 'get_c_niche']


class cell_network(object):
    """
    Collate the most basic neighborhood information.
    """

    def __init__(self,
                 adata,
                 cluster_key,
                 sample_key: Optional[str] = None,
                 sample=None,
                 ):
        if sample_key is not None:
            self.adata = adata[adata.obs[sample_key] == sample, :]
        else:
            self.adata = adata
        self._edge = _get_info_from_sample(self.adata, sample_id=sample, key='edges')
        self.cluster_key = cluster_key
        self.cluster = adata.obs[cluster_key]
        list_cluster = self.cluster.unique().tolist()
        new_list = [str(elem) for elem in list_cluster]
        self._cell_type_map = {v: i for i, v in enumerate(sorted(new_list))}
        self._species_of_clusters = len(self._cell_type_map)

    @property
    def cell_type(self):
        return self._cell_type_map.keys()

    @property
    def species_of_cell_type(self):
        return self._species_of_clusters


class cell_cell_interaction(cell_network):
    """
    Cell-cell interaction class
    """

    def __init__(self,
                 adata: anndata.AnnData,
                 cluster_key: str = 'clusters',
                 exclude: Union[str, dict] = None,
                 sample_key: Optional[str] = None,
                 sample=None,
                 ):

        super().__init__(adata, cluster_key, sample_key, sample)
        self.exclude = exclude

    def _randomize(self,
                   n_jobs: int,
                   num: int):
        adata = copy.deepcopy(self.adata)

        Iterator = Iterators(num,
                             adata,
                             self.cluster_key,
                             self._species_of_clusters,
                             self._cell_type_map)
        perms = Parallel(n_jobs=n_jobs)(delayed(_randomize_helper)(param[0], param[1], param[2], param[3])
                                        for param in Iterator)

        return perms

    def neighborhood_analysis(self,
                              n_jobs: int,
                              num: int,
                              method: str):

        @nb.njit()
        def enhance(matrix: np.ndarray, species: int):
            matrix_sum = np.sum(matrix, axis=0)
            for i in range(species):
                matrix[i, i] = matrix_sum[i] - matrix[i, i]
            for i in range(species):
                for j in range(i):
                    matrix[i, j] = matrix[i, j] / (matrix[i, i] + matrix[j, j] + 1)
                    matrix[j, i] = matrix[j, i] / (matrix[i, i] + matrix[j, j] + 1)
            for i in range(species):
                matrix[i, i] = 0

            return matrix

        mat_edge = _count_edge(self._edge, self._species_of_clusters, self._cell_type_map)
        perms = self._randomize(n_jobs=n_jobs, num=num)

        if method == 'included':
            pass
        elif method == 'excluded':
            mat_edge = enhance(mat_edge, self._species_of_clusters)
            species_list = [self._species_of_clusters] * num
            perms = Parallel(n_jobs=n_jobs)(delayed(enhance)(perm, spec) for perm, spec in zip(perms, species_list))
        perms = np.array(perms)
        zscore_array = (mat_edge - perms.mean(axis=0, dtype=np.float32)) / perms.std(axis=0, dtype=np.float32, ddof=1)
        zscore_df = pd.DataFrame(data=zscore_array, index=list(self._cell_type_map.keys()),
                                 columns=list(self._cell_type_map.keys()))

        return zscore_df

    def infiltration_analysis(self):

        # infiltration score
        mat_edge = _count_edge(self._edge, self._species_of_clusters, self._cell_type_map)
        mat_score = np.zeros((self.species_of_cell_type, self.species_of_cell_type))
        for index in range(self.species_of_cell_type):
            for column in range(self.species_of_cell_type):
                if index == column:
                    continue
                mat_score[index, column] = mat_edge[index, column] / min(mat_edge[index, index],
                                                                        mat_edge[column, column])

        pd_score = pd.DataFrame(mat_score, index=list(self._cell_type_map.keys()),
                                columns=list(self._cell_type_map.keys()))
        # self.adata.uns['SOAPy']['infiltration_score'] = pd_score

        # cell type number
        cluster_num = {}

        for c in self._cell_type_map.keys():
            cluster_num[c] = len(np.where(self.cluster == c)[0])

        return pd_score, cluster_num


class niche(cell_network):
    """
    Niche analysis. support multi-analysis.
    """

    def __init__(self,
                 adata: anndata.AnnData,
                 cluster_key: str = 'clusters',
                 sample_key: Optional[str] = None,
                 sample: str = None
                 ):

        super().__init__(adata, cluster_key, sample_key, sample=sample)
        self.sample = sample
        self.mat_neigh = None

    @property
    def i_niche(self):
        edge = self._edge
        mat_neigh = np.zeros((len(self.adata.obs_names), self._species_of_clusters))
        for index in edge.index:
            mat_neigh[int(edge.loc[index, 'point_1']), self._cell_type_map[edge.at[index, 'cluster_2']]] += 1
            mat_neigh[int(edge.loc[index, 'point_2']), self._cell_type_map[edge.at[index, 'cluster_1']]] += 1

        for index in range(len(self.adata.obs_names)):
            mat_neigh[index, self._cell_type_map[self.cluster[index]]] += 1

        pd_i_niche = pd.DataFrame(mat_neigh, index=self.adata.obs_names, columns=list(self.cell_type))

        for index in range(mat_neigh.shape[0]):
            if sum(mat_neigh[index, :]) != 0:
                mat_neigh[index, :] = mat_neigh[index, :] / sum(mat_neigh[index, :])
        self.mat_neigh = mat_neigh
        _add_info_from_sample(self.adata, sample_id=self.sample, keys='mat_neigh', add=mat_neigh)

        return pd_i_niche


class mult_sample_niche(object):
    def __init__(self,
                 adata: anndata.AnnData,
                 sample: Union[str, list, None],
                 celltype_key: str = 'clusters',
                 sample_key: Optional[str] = None,
                 ):
        list_i_niche = []
        list_barcode = []
        in_sample_cluster = []
        in_sample_list = []
        for sample_id in sample:
            i_niche_single = niche(adata, celltype_key, sample_key, sample=sample_id)
            list_i_niche.append(i_niche_single.i_niche)
            in_sample_cluster += i_niche_single.adata.obs[celltype_key].tolist()
            in_sample_list += i_niche_single.adata.obs[sample_key].tolist()
            list_barcode += i_niche_single.adata.obs_names.tolist()

        all_cluster = adata.obs[celltype_key].tolist()

        self.cluster = celltype_key
        self.in_sample_cluster = in_sample_cluster
        self.in_sample_list = in_sample_list
        self.barcode = list_barcode
        new_list = [str(elem) for elem in all_cluster]
        celltype_unique = np.unique(new_list)
        self._cell_type_map = {v: i for i, v in enumerate(sorted(celltype_unique))}
        self._species_of_clusters = len(self._cell_type_map)

        dict_num_spots = {}
        i_mult_niche = []
        for index, mat_niche in enumerate(list_i_niche):

            dict_num_spots[sample[index]] = mat_niche.shape[0]

            if self._species_of_clusters != mat_niche.shape[1]:
                mat_label = mat_niche.columns.tolist()
                label_add_mat = list(set(all_cluster) - set(mat_label))
                if len(label_add_mat) > 0:
                    for label in label_add_mat:
                        mat_niche[label] = [0.0 for i in range(mat_niche.shape[0])]

            if index == 0:
                i_mult_niche = copy.deepcopy(mat_niche)
                continue

            i_mult_niche = pd.concat([i_mult_niche, mat_niche])

        i_mult_niche = i_mult_niche.reindex(columns=np.unique(all_cluster), fill_value=0)
        self.mult_niche = i_mult_niche
        self.num_niche = dict_num_spots

    def mult_c_niche(self,
                     k_max,
                     sdbw: bool = True) -> pd.DataFrame:
        res, km = _best_k(self.mult_niche.values, k_max, sdbw)
        df_mult_niche = self.mult_niche
        # df_mult_niche = pd.DataFrame(data=self.mult_niche, columns=list(self._cell_type_map.keys()))

        df_mult_niche['sample'] = self.in_sample_list
        df_mult_niche['barcode'] = self.barcode
        df_mult_niche['celltype'] = self.in_sample_cluster
        df_mult_niche['C_niche'] = res

        return df_mult_niche


def neighborhood_analysis(
        adata: anndata.AnnData,
        method: Literal['excluded', 'include'] = 'excluded',
        cluster_key: str = 'clusters',
        sample_key: Optional[str] = None,
        sample: Union[int, str, list, None] = None,
        n_jobs: Optional[int] = None,
        n_iters: int = 1000,
        inplace: bool = True,
) -> anndata.AnnData:
    """
    Compute neighborhood enrichment Z-score by permutation test.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    method : str, optional
        'included': Z-scores of edges between two cell types were counted directly after randomization.
        'excluded': After randomization, remove self-connected edges between cells of the same type and calculate the z-score of edges between two cell types.
    cluster_key : str, optional
        The label of cluster in adata.obs.
    sample_key : str, optional
        The keyword of sample id in adata.obs.columns.
    sample: Union[int, str, list, None],
        The samples involved in calculating infiltration score.
    n_jobs : int, optional
        The maximum number of concurrently running jobs.
    n_iters : int, optional
        Number of rounds of random grouping for permutation tests.
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    - :attr:`anndata.AnnData.uns` ``['SOAPy']['include_method' or 'exclude_method']['dic_crd_poly']`` - neighborhood score

    """
    if sample is not None and sample_key is None:
        logg.error(f'Mult-sample niche analysis cannot be specified without a given sample key')
        raise ValueError
    if sample is None and sample_key is not None:
        sample = adata.obs[sample_key].unique().tolist()
        logg.info(f'Use all samples in the niche calculation')
    if not isinstance(sample, list):
        sample = [sample]

    adata = _check_adata_type(adata, 'spatial', inplace)

    for sample_id in sample:
        if sample_id is not None:
            bdata = adata[adata.obs[sample_key] == sample_id, :].copy()
            bdata.uns['SOAPy'] = {}
            bdata.uns['SOAPy'] = copy.deepcopy(adata.uns['SOAPy'][sample_id])
        else:
            bdata = adata
        new_cci = cell_cell_interaction(bdata, cluster_key)
        zscore = new_cci.neighborhood_analysis(n_jobs=n_jobs, num=n_iters, method=method)
        _add_info_from_sample(adata, sample_id=sample_id, keys=method + '_score', add=zscore)

    return adata


def infiltration_analysis(
        adata: anndata.AnnData,
        cluster_key: str = 'clusters',
        sample_key: Optional[str] = None,
        sample: Union[int, str, list, None] = None,
        inplace: bool = True,
) -> anndata.AnnData:
    """
    The infiltration score was calculated by the infiltration of non-parenchymal cells into parenchymal cells.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    cluster_key : str, optional
        The label of cluster in adata.obs.
    sample_key : str, optional
        The keyword of sample id in adata.obs.columns.
    sample: Union[int, str, list, None],
        The samples involved in calculating infiltration score.
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    - :attr:`anndata.AnnData.uns` ``['SOAPy']['infiltration score']['dic_crd_poly']`` - infiltration score

    """
    adata = _check_adata_type(adata, 'spatial', inplace)


    if sample is not None and sample_key is None:
        logg.error(f'Mult-sample niche analysis cannot be specified without a given sample key')
        raise ValueError
    if sample is None and sample_key is not None:
        sample = adata.obs[sample_key].unique().tolist()
        logg.info(f'Use all samples in the niche calculation')
    if not isinstance(sample, list):
        sample = [sample]

    for sample_id in sample:
        if sample_id is not None:
            bdata = adata[adata.obs[sample_key] == sample_id, :].copy()
            bdata.uns['SOAPy'] = {}
            bdata.uns['SOAPy'] = copy.deepcopy(adata.uns['SOAPy'][sample_id])
        else:
            bdata = adata

        new_cci = cell_cell_interaction(bdata, cluster_key)
        infiltration_score, cluster_num = new_cci.infiltration_analysis()

        _add_info_from_sample(adata,
                              sample_id=sample_id,
                              keys=['infiltration_analysis', 'cluster_number'],
                              add=[infiltration_score, cluster_num])
    _add_info_from_sample(adata, sample_id=None, keys='infiltration_sample', add=sample)

    return adata


# @decorator_helper
def get_c_niche(
        adata: anndata.AnnData,
        k_max: int,
        niche_key: str = 'C_niche',
        celltype_key: str = 'clusters',
        sample_key: Optional[str] = None,
        sample: Union[str, list] = None,
        sdbw: bool = False,
        inplace: bool = True,
) -> anndata.AnnData:
    """
    The C-niche is calculated using the cell type of the neighborhood.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    k_max:
        If sdbw is true, k_max is the maximum number of c-niche; If sdbw is false, it represents the number of c-niche
    niche_key : str, optional
        Add the keyword of niche in adata.obs.columns.
    celltype_key : str, optional
        The keyword of spot cluster in adata.obs.columns.
    sample_key : str, optional
        The keyword of sample id in adata.obs.columns.
    sample: Union[str, list, None],
        The samples involved in calculating the niche.
    sdbw : bool, optional
        Automated cluster number screening using sdbw.
    inplace : bool, optional
        Whether to change the original adata.

    Returns
    -------
    - :attr:`anndata.AnnData.obs` ``[niche]`` - c-niche of each spot
    - :attr:`anndata.AnnData.uns` ``[niche]`` - i-niche of each spot (cell type composition of neighbors)

    """
    adata = _check_adata_type(adata, 'spatial', inplace)

    if isinstance(sample, str):
        sample = [sample]
    if sample is not None and sample_key is None:
        logg.error(f'Mult-sample niche analysis cannot be specified without a given sample key')
        raise ValueError
    if sample is None and sample_key is not None:
        sample = list(adata.obs[sample_key].unique())
        logg.info(f'Use all samples in the niche calculation')

    assert celltype_key in adata.obs.columns

    niche_ = mult_sample_niche(adata=adata, sample=sample, celltype_key=celltype_key, sample_key=sample_key)
    niche_inf = niche_.mult_c_niche(k_max=k_max, sdbw=sdbw)
    _add_info_from_sample(adata=adata, sample_id=None, keys='niche', add=niche_inf)

    if sample is not None:
        adata.obs[niche_key] = None
        adata.obs.loc[np.isin(adata.obs[sample_key], sample), niche_key] = niche_inf['C_niche'].tolist()
    else:
        adata.obs[niche_key] = niche_inf['C_niche'].tolist()

    return adata




import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module

class Discriminator(nn.Module):
    def __init__(self, n_h):
        super(Discriminator, self).__init__()
        self.f_k = nn.Bilinear(n_h, n_h, 1)

        for m in self.modules():
            self.weights_init(m)

    def weights_init(self, m):
        if isinstance(m, nn.Bilinear):
            torch.nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0.0)

    def forward(self, c, h_pl, h_mi, s_bias1=None, s_bias2=None):
        c_x = c.expand_as(h_pl)  

        sc_1 = self.f_k(h_pl, c_x)
        sc_2 = self.f_k(h_mi, c_x)

        if s_bias1 is not None:
            sc_1 += s_bias1
        if s_bias2 is not None:
            sc_2 += s_bias2

        logits = torch.cat((sc_1, sc_2), 1)

        return logits
    
class AvgReadout(nn.Module):
    def __init__(self):
        super(AvgReadout, self).__init__()

    def forward(self, emb, mask=None):
        vsum = torch.mm(mask, emb)
        row_sum = torch.sum(mask, 1)
        row_sum = row_sum.expand((vsum.shape[1], row_sum.shape[0])).T
        global_emb = vsum / row_sum 
          
        return F.normalize(global_emb, p=2, dim=1) 
    
class Encoder(Module):
    def __init__(self, in_features, out_features, graph_neigh, dropout=0.0, act=F.relu):
        super(Encoder, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.graph_neigh = graph_neigh
        self.dropout = dropout
        self.act = act
        
        self.weight1 = Parameter(torch.FloatTensor(self.in_features, self.out_features))
        self.weight2 = Parameter(torch.FloatTensor(self.out_features, self.in_features))
        self.reset_parameters()
        
        self.disc = Discriminator(self.out_features)

        self.sigm = nn.Sigmoid()
        self.read = AvgReadout()
        
    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.weight1)
        torch.nn.init.xavier_uniform_(self.weight2)

    def forward(self, feat, feat_a, adj):
        z = F.dropout(feat, self.dropout, self.training)
        z = torch.mm(z, self.weight1)
        z = torch.mm(adj, z)
        
        hiden_emb = z
        
        h = torch.mm(z, self.weight2)
        h = torch.mm(adj, h)
        
        emb = self.act(z)
        
        z_a = F.dropout(feat_a, self.dropout, self.training)
        z_a = torch.mm(z_a, self.weight1)
        z_a = torch.mm(adj, z_a)
        emb_a = self.act(z_a)
        
        g = self.read(emb, self.graph_neigh) 
        g = self.sigm(g)  

        g_a = self.read(emb_a, self.graph_neigh)
        g_a = self.sigm(g_a)  

        ret = self.disc(g, emb, emb_a)  
        ret_a = self.disc(g_a, emb_a, emb) 
        
        return hiden_emb, h, ret, ret_a
    
class Encoder_sparse(Module):
    """
    Sparse version of Encoder
    """
    def __init__(self, in_features, out_features, graph_neigh, dropout=0.0, act=F.relu):
        super(Encoder_sparse, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.graph_neigh = graph_neigh
        self.dropout = dropout
        self.act = act
        
        self.weight1 = Parameter(torch.FloatTensor(self.in_features, self.out_features))
        self.weight2 = Parameter(torch.FloatTensor(self.out_features, self.in_features))
        self.reset_parameters()
        
        self.disc = Discriminator(self.out_features)

        self.sigm = nn.Sigmoid()
        self.read = AvgReadout()
        
    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.weight1)
        torch.nn.init.xavier_uniform_(self.weight2)

    def forward(self, feat, feat_a, adj):
        z = F.dropout(feat, self.dropout, self.training)
        z = torch.mm(z, self.weight1)
        z = torch.spmm(adj, z)
        
        hiden_emb = z
        
        h = torch.mm(z, self.weight2)
        h = torch.spmm(adj, h)
        
        emb = self.act(z)
        
        z_a = F.dropout(feat_a, self.dropout, self.training)
        z_a = torch.mm(z_a, self.weight1)
        z_a = torch.spmm(adj, z_a)
        emb_a = self.act(z_a)
         
        g = self.read(emb, self.graph_neigh)
        g = self.sigm(g)
        
        g_a = self.read(emb_a, self.graph_neigh)
        g_a =self.sigm(g_a)       
       
        ret = self.disc(g, emb, emb_a)  
        ret_a = self.disc(g_a, emb_a, emb)
        
        return hiden_emb, h, ret, ret_a     

class Encoder_sc(torch.nn.Module):
    def __init__(self, dim_input, dim_output, dropout=0.0, act=F.relu):
        super(Encoder_sc, self).__init__()
        self.dim_input = dim_input
        self.dim1 = 256
        self.dim2 = 64
        self.dim3 = 32
        self.act = act
        self.dropout = dropout
        
        #self.linear1 = torch.nn.Linear(self.dim_input, self.dim_output)
        #self.linear2 = torch.nn.Linear(self.dim_output, self.dim_input)
        
        #self.weight1_en = Parameter(torch.FloatTensor(self.dim_input, self.dim_output))
        #self.weight1_de = Parameter(torch.FloatTensor(self.dim_output, self.dim_input))
        
        self.weight1_en = Parameter(torch.FloatTensor(self.dim_input, self.dim1))
        self.weight2_en = Parameter(torch.FloatTensor(self.dim1, self.dim2))
        self.weight3_en = Parameter(torch.FloatTensor(self.dim2, self.dim3))
        
        self.weight1_de = Parameter(torch.FloatTensor(self.dim3, self.dim2))
        self.weight2_de = Parameter(torch.FloatTensor(self.dim2, self.dim1))
        self.weight3_de = Parameter(torch.FloatTensor(self.dim1, self.dim_input))
      
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.weight1_en)
        torch.nn.init.xavier_uniform_(self.weight1_de)
        
        torch.nn.init.xavier_uniform_(self.weight2_en)
        torch.nn.init.xavier_uniform_(self.weight2_de)
        
        torch.nn.init.xavier_uniform_(self.weight3_en)
        torch.nn.init.xavier_uniform_(self.weight3_de)
        
    def forward(self, x):
        x = F.dropout(x, self.dropout, self.training)
        
        #x = self.linear1(x)
        #x = self.linear2(x)
        
        #x = torch.mm(x, self.weight1_en)
        #x = torch.mm(x, self.weight1_de)
        
        x = torch.mm(x, self.weight1_en)
        x = torch.mm(x, self.weight2_en)
        x = torch.mm(x, self.weight3_en)
        
        x = torch.mm(x, self.weight1_de)
        x = torch.mm(x, self.weight2_de)
        x = torch.mm(x, self.weight3_de)
        
        return x
    
class Encoder_map(torch.nn.Module):
    def __init__(self, n_cell, n_spot):
        super(Encoder_map, self).__init__()
        self.n_cell = n_cell
        self.n_spot = n_spot
          
        self.M = Parameter(torch.FloatTensor(self.n_cell, self.n_spot))
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.M)
        
    def forward(self):
        x = self.M
        
        return x 



import random
import torch
import torch.nn as nn
from torch_geometric.nn import GCNConv, DeepGraphInfomax

import numpy as np
from scipy import sparse
import pickle


def sparse_mx_to_torch_edge_list(sparse_mx):
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    edge_list = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    return edge_list

def rep_dgi(
    n_h, 
    X, 
    A, 
    n_epoch=1000, 
    lr=0.001, 
    print_step=500, 
    torch_seed=None, 
    python_seed=None,
    numpy_seed=None,
    device=None
):
    # torch.set_deterministic(True)
    if not torch_seed is None:
        torch.manual_seed(torch_seed)
    if not python_seed is None:
        random.seed(python_seed)
    if not numpy_seed is None:
        np.random.seed(numpy_seed)


    n_f = X.shape[1]

    class Encoder(nn.Module):
        def __init__(self, in_channels, hidden_channels):
            super(Encoder, self).__init__()
            self.conv = GCNConv(in_channels, hidden_channels, cached=False)
            self.prelu = nn.PReLU(hidden_channels)
            self.conv2 = GCNConv(hidden_channels, hidden_channels, cached=False)
            self.prelu2 = nn.PReLU(hidden_channels)
        
        def forward(self, x, edge_index):
            x = self.conv(x, edge_index)
            x = self.prelu(x)
            x = self.conv2(x, edge_index)
            x = self.prelu2(x)
            return x
    
    def corruption(x, edge_index):
        return x[torch.randperm(x.size(0))], edge_index
    if device is None:
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    else:
        device = torch.device(device)
    model = DeepGraphInfomax(
        hidden_channels = n_h, encoder = Encoder(n_f, n_h),
        summary = lambda z, *args, **kwargs: torch.sigmoid(z.mean(dim=0)),
        corruption = corruption).to(device)
    
    X = torch.FloatTensor(X)
    X = X.to(device)
    edge_list = sparse_mx_to_torch_edge_list(A)
    edge_list = edge_list.to(device)

    optimiser = torch.optim.Adam(model.parameters(), lr=lr)

    def train():
        model.train()
        optimiser.zero_grad()
        pos_z, neg_z, summary = model(X, edge_list)
        loss = model.loss(pos_z, neg_z, summary)
        loss.backward()
        optimiser.step()
        return loss.item()

    for epoch in range(n_epoch):
        loss = train()
        if epoch % print_step == 0 or epoch+1 == n_epoch:
            print('Epoch: {:03d}, Loss: {:.4f}'.format(epoch, loss))
    
    z,_,_ = model(X, edge_list)
    return z.cpu().detach().numpy()

def main():
    f = open("./X.pkl", 'rb')
    X = pickle.load(f)
    X = X.toarray()
    f = open("./A.pkl", 'rb')
    A = pickle.load(f)
    z = rep_dgi(32, X, A)
    print(z.shape)

if __name__=="__main__":
    main()



import numpy as np
import gudhi
import networkx as nx
from sklearn.neighbors import kneighbors_graph
import matplotlib.pyplot as plt

def graph_alpha(pts, n_layer = 1, cut = np.inf, draw = False):
    
    # Get a graph from alpha shape
    pts_list = pts.tolist()
    n_node = len(pts_list)
    alpha_complex = gudhi.AlphaComplex(points=pts_list)
    simplex_tree = alpha_complex.create_simplex_tree(max_alpha_square=cut**2)
    skeleton = simplex_tree.get_skeleton(1)
    initial_graph = nx.Graph()
    initial_graph.add_nodes_from([i for i in range(n_node)])
    for s in skeleton:
        if len(s[0]) == 2:
            initial_graph.add_edge(s[0][0], s[0][1])
    # Extend the graph for the specified layers
    extended_graph = nx.Graph()
    extended_graph.add_nodes_from(initial_graph)
    extended_graph.add_edges_from(initial_graph.edges)
    if n_layer == 2:
        for i in range(n_node):
            for j in initial_graph.neighbors(i):
                for k in initial_graph.neighbors(j):
                    extended_graph.add_edge(i,k)
    elif n_layer == 3:
        for i in range(n_node):
            for j in initial_graph.neighbors(i):
                for k in initial_graph.neighbors(j):
                    for l in initial_graph.neighbors(k):
                        extended_graph.add_edge(i,l)
    if n_layer >= 4:
        print("Setting n_layer to greater than 3 may results in too large neighborhoods")    

    # Remove self edges
    for i in range(n_node):
        try:
            extended_graph.remove_edge(i,i)
        except:
            pass

    # Draw the graph
    if draw:
        pos_dict = {i:pts_list[i] for i in range(n_node)}
        nx.draw_networkx(extended_graph, pos=pos_dict, with_labels=False, node_size=1, node_color='dimgrey', edge_color='darkgrey')
        plt.axis('equal'); plt.axis('off'); plt.show()

    # Get the sparse adjacency matrix
    A = nx.to_scipy_sparse_matrix(extended_graph, format='csr')

    return A



import numpy as np
from scipy import sparse
from sklearn.neighbors import kneighbors_graph, radius_neighbors_graph
import networkx as nx
import matplotlib.pyplot as plt

def graph_knn(pts, k = 10, cut = np.inf, draw = False):
    # Assymetric knn graph
    A_knn = kneighbors_graph(pts, n_neighbors=k, mode='connectivity')
    # Make it symetric
    A_knn_sym = ((A_knn + A_knn.T).astype(bool)).astype(int)
    # Apply the cutoff
    if not np.isinf(cut):
        A_rn = radius_neighbors_graph(pts, radius=cut, mode='connectivity')
        A_knn_sym = A_knn_sym.multiply(A_rn)
    # Plot the graph
    if draw:
        G = nx.from_scipy_sparse_matrix(A_knn_sym)
        pts_list = list(pts)
        n_node = len(pts_list)
        pos_dict = {i:pts_list[i] for i in range(n_node)}
        nx.draw_networkx(G, pos=pos_dict, with_labels=False, node_size=10, node_color='dimgrey', edge_color='darkgrey')
        plt.axis('equal'); plt.axis('off'); plt.show()

    return A_knn_sym


"""
Cell communication analysis visualization related code.
Some source code is from stlearn.plotting.cci_plot (),we made some changes to suit our data structure.
"""
import copy
import anndata
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
import matplotlib.patches as patches
from matplotlib.colors import Colormap
import seaborn as sns
from scipy.stats.mstats import gmean
import networkx as nx
from anndata import AnnData
import math
from .utils import get_cmap, get_colors
from ._color import _get_palette
from ._chord import chordDiagram
from ._heatmap import _dotplot
from ..utils import _get_info_from_sample
import sys
from typing import Optional, Union, Tuple, Any, Literal
import logging

__all__ = [
    'show_ccc_chordplot',
    'show_ccc_netplot',
    'show_ccc_dotplot',
    'show_ccc_embedding'
]


def _get_lr_data(adata, sample_id, lr_type):

    data = _get_info_from_sample(adata=adata, sample_id=sample_id, key='celltype_comm_score')
    score = data[lr_type]['sig_celltype']
    title = f'cell-cell {lr_type} communication'

    return score, title


def _get_lr_matrix(
        adata: anndata.AnnData,
        lr_type: str,
        sample_id: Optional[str] = None,
        affinity_cutoff: Optional[float] = None,
        strength_cutoff: Optional[float] = None,
        lrs_name: Optional[list] = None,
        cts_name: Optional[list] = None,
        n_top_lrs: Optional[int] = None,
        n_top_cts: Optional[int] = None,
):

    def _3dto2d(data, len_ct, lr_pairs, cts):
        scores = np.zeros(shape=(len(lr_pairs), len_ct * len_ct))
        for index, name in enumerate(lr_pairs):
            p = data[index, :, :]
            k = 0
            for i in range(len_ct):
                for j in range(len_ct):
                    scores[index, k] = p[i, j]
                    k += 1

        scores = pd.DataFrame(scores, index=lr_pairs, columns=cts)
        return scores

    def _best_lr(
            strength: pd.DataFrame,
            affinity: pd.DataFrame,
            lrs_name: Optional[list] = None,
            cts_name: Optional[list] = None,
            affinity_cutoff: Optional[float] = None,
            strength_cutoff: Optional[float] = None,
            n_top_lrs: Optional[int] = None,
            n_top_cts: Optional[int] = None
        ):

        communication_intensity_p = np.where(affinity < affinity_cutoff, 1, 0)
        communication_intensity_s = np.where(strength > strength_cutoff, 1, 0)
        significance = communication_intensity_s & communication_intensity_p

        if lrs_name is None:
            sum_index_lr = np.sum(significance, axis=1)
            sort_index_lr = np.argsort(-sum_index_lr)
            if n_top_lrs is None or n_top_lrs > len(sort_index_lr):
                n_top_lrs = len(sort_index_lr)
            strength = strength.iloc[sort_index_lr[0: n_top_lrs], :]
            affinity = affinity.iloc[sort_index_lr[0: n_top_lrs], :]
        else:
            strength = strength.loc[lrs_name, :]
            affinity = affinity.loc[lrs_name, :]

        if cts_name is None:
            sum_index_cts = np.sum(significance, axis=0)
            sort_index_cts = np.argsort(-sum_index_cts)
            if n_top_cts is None or n_top_cts > len(sort_index_cts):
                n_top_cts = len(sort_index_cts)
            strength = strength.iloc[:, sort_index_cts[0:n_top_cts]]
            affinity = affinity.iloc[:, sort_index_cts[0:n_top_cts]]
        else:
            strength = strength.loc[:, cts_name]
            affinity = affinity.loc[:, cts_name]

        return strength, affinity

    data = _get_info_from_sample(adata=adata, sample_id=sample_id, key='celltype_comm_score')
    ct = data['celltype']
    if lr_type == 'contact':
        affinity = data['contact']['affinity']
        strength = data['contact']['strength']
        pairs = data['contact']['names']
    else:
        affinity = data['secretory']['affinity']
        strength = data['secretory']['strength']
        pairs = data['secretory']['names']

    cts = []
    for i in ct:
        for j in ct:
            cts.append(i + ',' + j)

    strength = _3dto2d(strength, len(ct), pairs, cts)
    affinity = _3dto2d(affinity, len(ct), pairs, cts)

    strength, affinity = _best_lr(strength=strength,
                                  affinity=affinity,
                                  affinity_cutoff=affinity_cutoff,
                                  lrs_name=lrs_name,
                                  cts_name=cts_name,
                                  strength_cutoff=strength_cutoff,
                                  n_top_lrs=n_top_lrs,
                                  n_top_cts=n_top_cts
                                  )

    return strength, affinity


def show_ccc_netplot(
        adata: AnnData,
        sample_id: Union[int, str] = None,
        lr_type: Literal['contact', 'secretory'] = 'contact',
        pos: dict = None,
        cmap: Optional[str] = None,
        font_size: int = 12,
        node_size_exp: int = 1,
        node_size_scaler: int = 1,
        min_counts: int = 0,
        ax: Axes = None,
        figsize: tuple = (8, 8),
        dpi: int = 100,
        show: bool = True,
        save: Optional[str] = None,
        **kwargs
) -> Optional[Axes]:
    """
    Visualize ligand-receptor interactions between two cell types in spatial omics data using network plot.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    ax : Axes
        A matplotlib axes object.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    lr_type: str
        The LR pair to visualise the cci network for. If None, will use spot
        cci counts across all LR pairs from adata.uns[f'lr_cci_use_label'].
    pos: dict
        Positions to draw each cell type, format as outputted from running
        networkx.circular_layout(graph). If not inputted will be generated.
    cmap: str
        Cmap to use when generating the cell colors, if not already specified by
        adata.uns[f'use_label_colors'].
    font_size: int
        Size of the cell type labels.
    node_size_scaler: float
        Scaler to multiply by node sizes to increase/decrease size.
    node_size_exp: int
        Increases difference between node sizes by this exponent.
    min_counts: int
        Minimum no. of LR interactions for connection to be drawn.
    kwargs : Any
        Other params of nx.draw_networkx()

    Returns
    -------
    pos: dict
        Dictionary of positions where the nodes are draw if return_pos is True, useful for consistent layouts.

    References
    ----------
    Pham, D. et al. Robust mapping of spatiotemporal trajectories and cell–cell interactions in healthy and diseased
        tissues. Nat Commun 14, 7739 (2023).

    """

    # Either plotting overall interactions, or just for a particular LR #
    int_df, title = _get_lr_data(adata, sample_id, lr_type)
    # Creating the interaction graph #
    all_set = int_df.index.values
    int_matrix = int_df.values
    graph = nx.MultiDiGraph()
    int_bool = int_matrix > min_counts
    int_matrix = int_matrix * int_bool
    for i, cell_A in enumerate(all_set):
        if cell_A not in graph:
            graph.add_node(cell_A)
        for j, cell_B in enumerate(all_set):
            if int_bool[i, j]:
                count = int_matrix[i, j]
                graph.add_edge(cell_A, cell_B, weight=count)

    # Determining graph layout, node sizes, & edge colours #
    if pos is None:
        pos = nx.circular_layout(graph)
    total = sum(sum(int_matrix))
    node_names = list(graph.nodes.keys())
    node_indices = [np.where(all_set == node_name)[0][0] for node_name in node_names]
    node_sizes = np.array(
        [
            (
                ((sum(int_matrix[i, :] + int_matrix[:, i]) - int_matrix[i, i]) / total)
                * 10000
                * node_size_scaler
            )
            ** (node_size_exp)
            for i in node_indices
        ]
    )
    node_sizes[node_sizes == 0] = 0.1

    edges = list(graph.edges.items())
    e_totals = []
    for i, edge in enumerate(edges):
        trans_i = np.where(all_set == edge[0][0])[0][0]
        receive_i = np.where(all_set == edge[0][1])[0][0]
        e_total = (
            sum(list(int_matrix[trans_i, :]) + list(int_matrix[:, receive_i]))
            - int_matrix[trans_i, receive_i]
        )
        e_totals.append(e_total)
    edge_weights = [edge[1]["weight"] / e_totals[i] for i, edge in enumerate(edges)]

    # Determining node colors #
    nodes = np.unique(list(graph.nodes.keys()))
    node_colors = _get_palette(nodes, palette=cmap)
    node_colors = list(node_colors.values())
    if not np.all(np.array(node_names) == nodes):
        nodes_indices = [np.where(nodes == node)[0][0] for node in node_names]
        node_colors = np.array(node_colors)[nodes_indices]

    #### Drawing the graph #####
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
    else:
        fig = ax.get_figure()

    # Adding in the self-loops #
    z = 55
    for i, edge in enumerate(edges):
        cell_type = edge[0][0]
        if cell_type != edge[0][1]:
            continue
        x, y = pos[cell_type]
        angle = math.degrees(math.atan(y / x))
        if x > 0:
            angle = angle + 180
        arc = patches.Arc(
            xy=(x, y),
            width=0.3,
            height=0.025,
            lw=5,
            ec=plt.cm.get_cmap("Blues")(edge_weights[i]),
            angle=angle,
            theta1=z,
            theta2=360 - z,
        )
        ax.add_patch(arc)

    # Drawing the main components of the graph #
    nx.draw_networkx(
        graph,
        pos,
        node_size=node_sizes,
        node_color=node_colors,
        arrowstyle="->",
        arrowsize=50,
        width=5,
        font_size=font_size,
        font_weight="bold",
        edge_color=edge_weights,
        ax=ax,
        **kwargs
    )
    fig.suptitle(title, fontsize=30)
    plt.tight_layout()

    if save is not None:
        plt.savefig(save)
    if show:
        plt.show()
    else:
        return ax


def show_ccc_chordplot(
        adata: AnnData,
        sample: Union[str, list] = None,
        lr_type: Literal['contact', 'secretory'] = 'contact',
        min_ints: int = 2,
        n_top_ccis: int = 10,
        cmap: str = None,
        label_size: int = 10,
        label_rotation: float = 0,
        ax: Axes = None,
        figsize: tuple = (8, 8),
        dpi: int = 100,
        show: bool = True,
        save: Optional[str] = None,
) -> Optional[Axes]:
    """
    Visualize ligand-receptor interactions between two cell types in spatial omics data using chord plot.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    ax : Axes
        A matplotlib axes object.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    sample : Union[str, list]
        Name of sample.
    lr_type : Literal['contact', 'secretory']
        Types of ligand-receptor pairs.
    min_ints : int
        Minimum number of interactions for a connection to be considered.
    n_top_ccis : int
        Maximum number of top ligand-receptor pairs to display.
    cmap : str
        Colormap for visualizing cell types.
    label_size : int
        Font size for node labels.
    label_rotation : float
        Rotation angle for node labels.

    Returns
    -------
    ax : matplotlib.axes.Axes
        The axes object containing the chord plot.

    References
    ----------
    Pham, D. et al. Robust mapping of spatiotemporal trajectories and cell–cell interactions in healthy and diseased
        tissues. Nat Commun 14, 7739 (2023).
    """
    # Either plotting overall interactions, or just for a particular LR #
    int_df, title = _get_lr_data(adata, sample, lr_type)

    int_df = int_df.transpose()

    flux = int_df.values
    total_ints = flux.sum(axis=1) + flux.sum(axis=0) - flux.diagonal()
    keep = np.where(total_ints > min_ints)[0]
    # Limit of 10 for good display #
    if len(keep) > n_top_ccis:
        keep = np.argsort(-total_ints)[0:n_top_ccis]
    # Filter any with all zeros after filtering #
    all_zero = np.array(
        [np.all(np.logical_and(flux[i, keep] == 0, flux[keep, i] == 0)) for i in keep]
    )
    keep = keep[all_zero == False]
    if len(keep) == 0:  # If we don't keep anything, warn the user
        return

    flux = flux[:, keep]
    flux = flux[keep, :].astype(float)

    # Add pseudocount to row/column which has all zeros for the incoming
    # so can make the connection between the two
    for i in range(flux.shape[0]):
        if np.all(flux[i, :] == 0):
            flux[i, flux[:, i] > 0] += sys.float_info.min
        elif np.all(flux[:, i] == 0):
            flux[flux[i, :] > 0, i] += sys.float_info.min

    cell_names = int_df.index.values.astype(str)[keep]
    nodes = cell_names

    # Retrieving colors of cell types #
    cluster = np.unique(int_df.index)
    colors = _get_palette(cluster, palette=cmap)
    colors = list(colors.values())

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
    else:
        fig = ax.get_figure()

    nodePos = chordDiagram(flux, ax, lim=1.25, colors=colors)
    ax.axis("off")
    prop = dict(fontsize=label_size, ha="center", va="center")
    for i in range(len(cell_names)):
        x, y = nodePos[i][0:2]
        rotation = nodePos[i][2]
        # Prevent text going upside down at certain rotations
        if (90 > rotation > 18 and label_rotation != 0) or (120 > rotation > 90):
            label_rotation_ = -label_rotation
        else:
            label_rotation_ = label_rotation
        ax.text(
            x, y, nodes[i], rotation=nodePos[i][2] + label_rotation_, **prop
        )  # size=10,
    fig.suptitle(title, fontsize=20)

    if save is not None:
        plt.savefig(save)
    if show:
        plt.show()
    else:
        return ax


def show_ccc_dotplot(
        adata: AnnData,
        sample: Union[str, list] = None,
        lr_type: Literal['contact', 'secretory'] = 'contact',
        affinity_cutoff: float = 0.05,
        strength_cutoff: float = 2.0,
        lrs_name: Optional[list] = None,
        cts_name: Optional[list] = None,
        n_top_lrs: Optional[int] = 10,
        n_top_cts: Optional[int] = 15,
        figsize: Tuple[Any, Any] = (7, 5),
        cmap: Union[Colormap, str] = "Spectral_r",
        strength_min: Optional[float] = None,
        strength_max: Optional[float] = None,
        size: int = 10,
        sig_color: str = '#000000',
        ax: Axes = None,
        show: bool = False,
        save: Optional[str] = None,
        **kwargs
) -> Optional[Axes]:
    """
    Dotplot for cell communication analysis are presented, either specifying the cell type pairs and ligand-receptor
    pairs to be presented or showing the most significant cell type pairs and ligand-receptor pairs after ranking by
    significance level.

    Parameters
    ----------
    adata : anndata.Anndata
        An AnnData object containing spatial omics data and spatial information.
    sample : Union[str, list]
        Name of sample.
    lr_type : Literal['contact', 'secretory']
        Types of ligand-receptor pairs.
    affinity_cutoff : float
        The threshold at which affinity was significant, and affinity below this threshold were considered significant.
    strength_cutoff : float
        The threshold at which strength was significant, and strength above this threshold were considered significant.
    lrs_name : Optional[list]
        Ligand-receptor pairs that need to be show.
        e.g. ['Endothelial cell,Macrophage', 'Macrophage,Endothelial cell']
    cts_name : Optional[list]
        cell type pairs that need to be show.
        e.g. ['FGF2:FGFR1', 'WNT3A:FZD7&LRP5']
    n_top_lrs : int
        Maximum number of ligand-receptor pairs presented, ordered according to the number of cell type pairs for which
        a ligand-receptor pair was significant.
    n_top_cts : int
        The maximum number of cell type pairs shown, ordered according to the number of significant ligand-receptor
        pairs for a cell type pair.
    figsize : Tuple[Any, Any]
        The size of format figure.(width, height)
    cmap : Union[Colormap, str]
        Color map of dotplot.
    strength_min, strength_max : Optional[float]
        The extreme value of the strength that is displayed, the value that is not in the range will be displayed in the
         same color as the corresponding extreme value.
    size : int
        Size of the points in figure.
    sig_color : str
        Points that were considered significant were highlighted by the color.
    ax : matplotlib.figure.Axes
        A matplotlib.axes object.
    show : str
        Show this plot.
    save : Optional[str]
        The path where the image is stored.
    kwargs: Any, optional
        Other params of ax.scatter()

    Returns
    -------
        If show==False, return Axes
    """
    strength, affinity = \
        _get_lr_matrix(adata=adata,
                       affinity_cutoff=affinity_cutoff,
                       strength_cutoff=strength_cutoff,
                       lrs_name=lrs_name,
                       cts_name=cts_name,
                       n_top_lrs=n_top_lrs,
                       n_top_cts=n_top_cts,
                       sample_id=sample,
                       lr_type=lr_type
                       )

    # heatmap
    ax = _dotplot(
        scores=strength,
        pvalue=affinity,
        s_cutoff=strength_cutoff,
        p_cutoff=affinity_cutoff,
        vmin=strength_min,
        vmax=strength_max,
        ax=ax,
        cmap=cmap,
        sig_color=sig_color,
        figsize=figsize,
        s=size,
        **kwargs
    )
    ax.set_title(f'{lr_type} lr pairs')
    ax.set_ylabel('LR-pair')
    ax.set_xlabel('cell type')

    legend_elements = [
        plt.scatter([], [], marker='o', color='silver', edgecolor='#FFFFFF', s=size, label='> 0.5'),
        plt.scatter([], [], marker='o', color='silver', edgecolor='#FFFFFF', s=size * 2, label='> 0.1'),
        plt.scatter([], [], marker='o', color='silver', edgecolor='#FFFFFF', s=size * 3, label='> 0.05'),
        plt.scatter([], [], marker='o', color='silver', edgecolor='#FFFFFF', s=size * 4, label='< 0.05')
        ]
    plt.legend(handles=legend_elements, bbox_to_anchor=(1.2, 0.0), title='P value')

    plt.tight_layout()

    if save is not None:
        plt.savefig(save)
    if show:
        plt.show()
    else:
        return ax


def show_ccc_embedding(
        adata: anndata.AnnData,
        ligand: Union[str, list],
        receptor: Union[str, list],
        ligand_clusters: Union[list, str, int],
        receptor_clusters: Union[list, str, int],
        cluster_key: str = 'cell_type',
        ligand_region: Tuple[Any, Any] = None,
        receptor_region: Tuple[Any, Any] = None,
        ligand_color: Union[Colormap, str] = None,
        receptor_color: Union[Colormap, str] = None,
        row_region: Tuple[Any, Any] = None,
        col_region: Tuple[Any, Any] = None,
        agg_method: Literal['mean', 'min', 'gmean'] = 'gmean',
        title: Optional[str] = None,
        figsize: Tuple[Any, Any] = None,
        dpi: int = 100,
        ax: Optional[Axes] = None,
        size: Optional[float] = None,
        obsm_spatial_label: str = 'spatial',
        show: bool = True,
        save: Optional[str] = None,
        cbar_kwargs: Optional[dict] = {},
        **kwargs
) -> Optional[Axes]:
    """
    To map the distribution of ligands and receptors expression of different cell type in a situ map of tissue.

    Parameters
    ----------
    adata : anndata.AnnData
        An AnnData object containing spatial omics data and spatial information.
    figsize : Tuple[float, float], optional
        (Width, height) of the figure.
    dpi : float, optional
        The resolution of the figure.
    title : str, optional
        The title of shown figure.
    ax : Axes
        A matplotlib axes object.
    show : bool
        Show the plot, do not return axis.
    save : Union[str, PathLike], optional
        The path where the image is stored.
    ligand : Union[str, list]
        Ligand name or a list containing ligand subunits
    receptor : Union[str, list]
        Receptor name or a list containing receptor subunits
    ligand_clusters : Union[list, str, int]
        The cell type clusters of the ligand shown, cannot be duplicated with the clusters of the receptor,
        otherwise overlap would occur.
    receptor_clusters : Union[list, str, int]
        The cell type clusters of the receptor shown, cannot be duplicated with the clusters of the ligand,
        otherwise overlap would occur.
    cluster_key : str
        The label of cluster in adata.obs.
    ligand_region : Tuple[Any, Any]
        The range of values of the ligand. (ligand_min, ligand_max)
    receptor_region : Tuple[Any, Any]
        The range of values of the receptor. (receptor_min, receptor_max)
    ligand_color : Union[Colormap, str]
        The color of the ligand.
    receptor_color : Union[Colormap, str]
        The color of the receptor.
    row_region : Tuple[Any, Any]
        The row coordinate threshold of the region in the situ map is displayed, and the expression of the whole
        situ map is displayed by default. (row_low, row_high)
    col_region : Tuple[Any, Any]
        The row coordinate threshold of the region in the situ map is displayed, and the expression of the whole
        situ map is displayed by default. (row_low, row_high)
    agg_method : str
        Integrated subunit to calculate ligand or receptor.
        'min': ligand or receptor = min(sub1, sub2, sub3, ...)
        'mean': ligand or receptor = mean(sub1, sub2, sub3, ...)
        'gmean': geometrical mean, ligand or receptor = np.gmean(sub1, sub2, sub3, ...)
    size : Optional[float]
        Size of the points in figure.
    obsm_spatial_label : str
        The key of spatial coordinates in adata.obsm.
    cbar_kwargs : Optional[dict]
        Other parameters in plt.colorbar().
    kwargs : Any, optional
        Other params of ax.scatter()

    Returns
    -------
        If show==False, return Axes
    """

    def subunit_merging(exp, agg_method):
        if agg_method == 'min':
            update_ = np.min(exp, axis=1)
        elif agg_method == 'mean':
            update_ = np.mean(exp, axis=1)
        elif agg_method == 'gmean':
            update_ = gmean(exp, axis=1)
        else:

            ValueError("{} is not a valid agg_method".format(agg_method))

        return update_

    def in_region(data, up, down, method):
        up_index = np.where(data > up, True, False)
        down_index = np.where(data < down, True, False)
        if method == 'drop':
            index_all = up_index | down_index
            data = np.array([-1 if index_all[i] else x for i, x in enumerate(data)])
            return data
        else:
            data[up_index] = up
            data[down_index] = down
            return data

    adata_ligand = copy.deepcopy(adata)
    adata_receptor = copy.deepcopy(adata)
    clusters = adata.obs[cluster_key].tolist()

    if not isinstance(ligand_clusters, list):
        ligand_clusters = [ligand_clusters]

    if not isinstance(receptor_clusters, list):
        receptor_clusters = [receptor_clusters]

    if np.any(np.isin(ligand_clusters, receptor_clusters)):
        logging.error(
            'The ligand clusters is duplicated with the receptor clusters. Duplicate categories\' clusters will be overwritten.'
            'Please check it.'
        )

    index_ligand = np.isin(clusters, ligand_clusters)
    adata_ligand = adata_ligand[index_ligand, :]
    index_receptor = np.isin(clusters, receptor_clusters)
    adata_receptor = adata_receptor[index_receptor, :]

    ligand_x, ligand_y = adata_ligand.obsm[obsm_spatial_label].T
    receptor_x, receptor_y = adata_receptor.obsm[obsm_spatial_label].T
    if isinstance(ligand, list):
        ligand_exp = adata_ligand[:, ligand].X.toarray()
        ligand_exp = subunit_merging(ligand_exp, agg_method)
    else:
        ligand_exp = adata_ligand[:, ligand].X.toarray().T[0]
    if isinstance(receptor, list):
        receptor_exp = adata_receptor[:, receptor].X.toarray()
        receptor_exp = subunit_merging(receptor_exp, agg_method)
    else:
        receptor_exp = adata_receptor[:, receptor].X.toarray().T[0]

    if col_region is not None:
        ligand_x = in_region(ligand_x, up=col_region[1], down=col_region[0], method='drop')
        receptor_x = in_region(receptor_x, up=col_region[1], down=col_region[0], method='drop')
    if row_region is not None:
        ligand_y = in_region(ligand_y, up=row_region[1], down=row_region[0], method='drop')
        receptor_y = in_region(receptor_y, up=row_region[1], down=row_region[0], method='drop')

    if ligand_region is not None:
        ligand_exp = in_region(ligand_exp, up=ligand_region[1], down=ligand_region[0], method='no drop')
    if receptor_region is not None:
        receptor_exp = in_region(receptor_exp, up=receptor_region[1], down=receptor_region[0], method='no drop')

    index_ = []
    for i in range(len(ligand_x)):
        if np.isin(-1, [ligand_x[i], ligand_y[i]]):
            index_.append(False)
        else:
            index_.append(True)

    ligand_x = ligand_x[index_]
    ligand_y = ligand_y[index_]
    ligand_exp = ligand_exp[index_]

    index_ = []
    for i in range(len(receptor_x)):
        if np.isin(-1, [receptor_x[i], receptor_y[i]]):
            index_.append(False)
        else:
            index_.append(True)

    receptor_x = receptor_x[index_]
    receptor_y = receptor_y[index_]
    receptor_exp = receptor_exp[index_]

    if size is None:
        size = 120000 / adata.shape[0]

    if ax is None:
        if figsize is None:
            figsize = (10, 8)
        fig = plt.figure(figsize=figsize, dpi=dpi)
        ax = fig.add_subplot(111)
    else:
        fig = ax.get_figure()

    if ligand_color is None:
        # ligand_color = ListedColormap(sns.color_palette("YlOrBr", as_cmap=True))
        ligand_color = sns.color_palette("YlOrBr", as_cmap=True)
    if receptor_color is None:
        # receptor_color = ListedColormap(sns.color_palette("Blues", as_cmap=True))
        from matplotlib.colors import LinearSegmentedColormap
        receptor_color = sns.color_palette("Blues", as_cmap=True)
        receptor_color = LinearSegmentedColormap.from_list(
            "custom_blues", receptor_color(np.linspace(0.3, 1.0, 256))
        )

    ax_0 = ax.scatter(
        ligand_x,
        ligand_y,
        s=size,
        c=ligand_exp,
        cmap=ligand_color,
        **kwargs
    )

    ax_1 = ax.scatter(
        receptor_x,
        receptor_y,
        s=size,
        c=receptor_exp,
        cmap=receptor_color,
        **kwargs
    )

    ax.set_xlabel(obsm_spatial_label+'1')
    ax.set_ylabel(obsm_spatial_label+'2')

    ax.set_yticks([])
    ax.set_xticks([])

    cbar1 = plt.colorbar(ax_0, ax=ax, orientation='vertical', pad=0.02, **cbar_kwargs)
    cbar2 = plt.colorbar(ax_1, ax=ax, orientation='vertical', pad=0.02, **cbar_kwargs)

    cbar1.ax.set_title('ligand', fontsize=12)
    cbar2.ax.set_title('receptor', fontsize=12)

    cbar1.ax.set_position([0.8, 0.55, 0.06, 0.35])  # 上方的颜色棒
    cbar2.ax.set_position([0.8, 0.1, 0.06, 0.35])  # 下方的颜色棒

    ax.autoscale_view()

    if title is not None:
        title = f'{ligand_clusters}:{ligand},\n {receptor_clusters}:{receptor}'

    fig.suptitle(title, fontsize=20)

    if save is not None:
        plt.savefig(save)

    if show:
        plt.show()
    else:
        return ax


import numpy as np
import pandas as pd
import seaborn as sns
from typing import Optional, Union
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from ..tl._tensor import TensorDecomposition

__all__ = [
    'show_factor_matrix_in_CP_tensor',
    'show_factor_matrix_in_tucker_tensor',
    'show_proportion_in_CP',
    'show_proportion_in_tucker'
]


def show_factor_matrix_in_CP_tensor(
        tensor: TensorDecomposition,
        factor_name: str,
        **kwargs
):
    """
    Visualizes a factor in the CP decomposition tensor.

    Parameters
    ----------
    tensor : TensorDecomposition
        The TensorDecomposition object containing the CP decomposition.
    factor_name : str
        The name of the factor to visualize.
    **kwargs : dict
        Additional keyword arguments for seaborn.clustermap.

    Returns
    -------
    None
    """

    tensor_object = tensor
    rank = tensor_object.CP['rank']
    factor = tensor_object.CP['factors'][factor_name]
    dict_label = tensor_object.dict_sup[factor_name]
    _show_factor(factor, dict_label, rank, factor_name, **kwargs)


def show_factor_matrix_in_tucker_tensor(
        tensor: TensorDecomposition,
        factor_name: str,
        **kwargs
):
    """
    Visualizes a factor in the Tucker decomposition tensor.

    Parameters
    ----------
    tensor : TensorDecomposition
        The TensorDecomposition object containing the Tucker decomposition.
    factor_name : str
        The name of the factor to visualize.
    **kwargs : dict
        Additional keyword arguments for seaborn.clustermap.

    Returns
    -------
    None
    """

    tensor_object = tensor
    factor_rank = tensor_object.tucker['rank'][factor_name]
    factor = tensor_object.tucker['factors'][factor_name]
    dict_label = tensor_object.dict_sup[factor_name]
    _show_factor(factor, dict_label, factor_rank, factor_name, **kwargs)


def show_proportion_in_CP(
        tensor: TensorDecomposition,
        module: int,
        figsize: Optional[tuple] = None,
        factor_name: Union[list, str, None] = None,
        max_key: Union[None, int, list] = None,
        order: bool = True,
        show: bool = True,
        save: Optional[str] = None,
        **kwargs
):
    """
    Visualizes the proportion in a CP decomposition.

    Parameters
    ----------
    tensor : TensorDecomposition
        The TensorDecomposition object containing the CP decomposition.
    module : int
        The module to visualize.
    figsize : tuple, optional
        Figure size.
    factor_name : Union[list, str, None], optional
        The name of the factor to visualize.
    max_key : Union[None, int, list], optional
        Maximum number of keys to show.
    order : bool, optional
        Whether to order the keys.
    show : bool, optional
        Whether to display the figure.
    save : str, optional
        Filepath to save the figure.
    **kwargs : dict
        Additional keyword arguments for seaborn.barplot.

    Returns
    -------
    None or list of Axes
        If show is False, returns the list of Axes.
    """

    tensor_object = tensor
    factors = tensor_object.CP['factors']
    dict_sup = tensor_object.dict_sup
    if factor_name is None:
        factor_name = tensor_object.factor_name

    if not isinstance(factor_name, list):
        factor_name = [factor_name]
    if not isinstance(max_key, list):
        max_key = [max_key]*len(factor_name)
    if figsize is None:
        figsize = (6 * len(factor_name), 6)
    fig, axes = plt.subplots(1, len(factors), figsize=figsize)
    for index, name in enumerate(factor_name):
        _show_proportion(factor=factors[name],
                         module=module,
                         dict_label=dict_sup[name],
                         factor_name=name,
                         ax=axes[index],
                         max_key=max_key[index],
                         order=order,
                         **kwargs)
    fig.suptitle('module '+str(module))
    fig.tight_layout()
    if save:
        fig.savefig(save)
    if show:
        fig.show()
    else:
        return axes


def show_proportion_in_tucker(
        tensor: TensorDecomposition,
        module: list,
        figsize: Optional[tuple] = None,
        factor_name: Union[list, str, None] = None,
        max_key: Union[None, int, list] = None,
        order: bool = True,
        show: bool = True,
        save: Optional[str] = None,
        **kwargs
):
    """
    Visualizes the proportion in a Tucker decomposition.

    Parameters
    ----------
    tensor : TensorDecomposition
        The TensorDecomposition object containing the Tucker decomposition.
    module : list
        List of modules to visualize.
    figsize : tuple, optional
        Figure size.
    factor_name : Union[list, str, None], optional
        The name of the factor to visualize.
    max_key : Union[None, int, list], optional
        Maximum number of keys to show.
    order : bool, optional
        Whether to order the keys.
    show : bool, optional
        Whether to display the figure.
    save : str, optional
        Filepath to save the figure.
    **kwargs : dict
        Additional keyword arguments for seaborn.barplot.

    Returns
    -------
    None or list of Axes
        If show is False, returns the list of Axes.
    """
    tensor_object = tensor
    factors = tensor_object.CP['factors']
    dict_sup = tensor_object.dict_sup
    if factor_name is None:
        factor_name = tensor_object.factor_name

    if not isinstance(factor_name, list):
        factor_name = [factor_name]
    if not isinstance(max_key, list):
        max_key = [max_key] * len(factor_name)
    if figsize is None:
        figsize = (6 * len(factor_name), 6)
    fig, axes = plt.subplots(1, len(factors), figsize=figsize)
    for index, name in enumerate(factor_name):
        _show_proportion(factors[name],
                         module[index],
                         dict_sup[name], name, ax=axes[index], max_key=max_key[index],
                         order=order, **kwargs)

    plt.suptitle('module '+str(module))
    plt.show()
    if save:
        fig.savefig(save)
    if show:
        fig.show()
    else:
        return axes


def _show_factor(factor: np.ndarray,
                 dict_label: dict,
                 rank: int,
                 factor_name: str,
                 **kwargs
                 ):
    """
    Display a factor using seaborn.clustermap.

    Parameters
    ----------
    factor : np.ndarray
        The factor to display.
    dict_label : dict
        Dictionary of labels.
    rank : int
        Rank of the factor.
    factor_name : str
        Name of the factor.
    **kwargs : dict
        Additional keyword arguments for seaborn.clustermap.

    Returns
    -------
    None
    """

    index_name = list(dict_label.keys())
    col_name = ['factor ' + str(i) for i in range(rank)]
    tensor_df = pd.DataFrame(factor, index=index_name, columns=col_name)
    sns.clustermap(tensor_df, col_cluster=False, **kwargs)
    plt.xticks(rotation=45)
    plt.suptitle(factor_name)
    plt.show()


def _show_proportion(factor: np.ndarray,
                     module,
                     dict_label: dict,
                     factor_name,
                     ax,
                     max_key: Optional[int] = None,
                     order: bool = True,
                     **kwargs
                     ):
    """
    Display the proportion using seaborn.barplot.

    Parameters
    ----------
    factor : np.ndarray
        The factor to display.
    module : int
        The module to visualize.
    dict_label : dict
        Dictionary of labels.
    factor_name : str
        Name of the factor.
    ax : matplotlib.axes._subplots.AxesSubplot
        Matplotlib axes for plotting.
    max_key : Optional[int], optional
        Maximum number of keys to show.
    order : bool, optional
        Whether to order the keys.
    **kwargs : dict
        Additional keyword arguments for seaborn.barplot.

    Returns
    -------
    None
    """

    x = np.array(list(dict_label.keys()))
    y = np.array(factor[:, module])
    if order:
        index = np.argsort(y)[::-1]
        x = x[index]
        y = y[index]
    if max_key is not None:
        if len(x) > max_key:
            x = x[: max_key]
            y = y[: max_key]
    sns.barplot(x=x, y=y, ax=ax, **kwargs)
    ax.set_xticklabels(x, rotation=20, fontsize=10)
    ax.set_title(factor_name)